"""
This module contains the base classes for visual stimuli.
"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import tempfile
import os
import copy
import random
import itertools

import pygame
try:
    import OpenGL.GLU as oglu
    import OpenGL.GL as ogl
except ImportError:
    oglu = None
    ogl = None
try:
    import numpy as np
except ImportError:
    np = None

from . import defaults
from .. import _internals
from ._stimulus import Stimulus
from .. import misc
from ..misc import geometry, unicode2byte
from ..misc._timer import get_time

random.seed()


class _LaminaPanelSurface(object):
    """A class implementing an OpenGL surface."""

    # The following code is based on part of the Lamina module by David Keeney
    # (https://pypi.org/project/Lamina/) with some modifications to fit it
    # into expyriment (e.g. positioning)
    def __init__(self, surface, quadDims=(-1, 1, 1, 1),
                 position=(0, 0)):
        """Initialize new instance.

        Parameters
        ----------
        surface : pygame.Surface or numpy.array object
            surface to convert
        quadDims : (int,int), optional
        position : (int,int), optional

        """

        if isinstance(surface, pygame.Surface):
            # Fix for uneven surface sizes
            surface_size = surface.get_size()
            if surface_size[0] % 2 == 1:
                rect = pygame.Rect((0, 0), surface_size)
                s = pygame.surface.Surface((surface_size[0] + 1, surface_size[1]), pygame.SRCALPHA).convert_alpha()
                s.blit(surface, rect)
                surface = s
            surface_size = surface.get_size()
            if surface_size[1] % 2 == 1:
                rect = pygame.Rect((0, 0), surface_size)
                s = pygame.surface.Surface((surface_size[0], surface_size[1] + 1), pygame.SRCALPHA).convert_alpha()
                s.blit(surface, rect)
                surface = s
            self._winsize = surface.get_size()
        else:
            self._winsize = (len(surface[0]), len(surface))
        self._txtr = Visual._load_texture(surface)
        self._position = position
        left, top, width, height = quadDims
        right, bottom = left + width, top - height
        self._qdims = quadDims
        self.dims = ((left, top, 0), (right, top, 0),
                     (right, bottom, 0), (left, bottom, 0))
        self.refresh_position()

    def __del__(self):
        """Call glDeleteTextures when deconstruction the object."""

        if getattr(self, '_txtr', None) is not None:
            try:
                ogl.glDeleteTextures([self._txtr])
            except:
                pass

    def convertMousePos(self, pos):
        """Convert 2d pixel mouse pos to 2d gl units.

        Parameters
        ----------
        pos : (int, int)
            position of mouse

        """

        x0, y0 = pos
        x = int(x0 / self._winsize[0] * self._qdims[2] + self._qdims[0])
        y = int(y0 / self._winsize[1] * self._qdims[3] + self._qdims[1])
        return x, y

    def refresh_position(self):
        """Recalc where in modelspace quad needs to be to fill screen."""

        screensize = pygame.display.get_surface().get_size()
        bottomleft = oglu.gluUnProject(screensize[0] // 2 - \
                                       int(misc.round(self._winsize[0] / 2)) + \
                                       self._position[0],
                                       screensize[1] // 2 - \
                                       self._winsize[1] // 2 + \
                                       self._position[1], 0)
        bottomright = oglu.gluUnProject(screensize[0] // 2 + \
                                        self._winsize[0] // 2 + \
                                        self._position[0],
                                        screensize[1] // 2 - \
                                        self._winsize[1] // 2 + \
                                        self._position[1], 0)
        topleft = oglu.gluUnProject(screensize[0] // 2 - \
                                    int(misc.round(self._winsize[0] / 2)) + \
                                    self._position[0],
                                    screensize[1] // 2 + \
                                    int(misc.round(self._winsize[1] / 2)) + \
                                    self._position[1], 0)
        topright = oglu.gluUnProject(screensize[0] // 2 + \
                                     self._winsize[0] // 2 + \
                                     self._position[0],
                                     screensize[1] // 2 + \
                                     int(misc.round(self._winsize[1] / 2)) + \
                                     self._position[1], 0)


        self.dims = topleft, topright, bottomright, bottomleft
        width = topright[0] - topleft[0]
        height = topright[1] - bottomright[1]
        self._qdims = topleft[0], topleft[1], width, height

    def display(self):
        """Draw surface to a quad."""

        ogl.glEnable(ogl.GL_BLEND)
        ogl.glBlendFunc(ogl.GL_SRC_ALPHA, ogl.GL_ONE_MINUS_SRC_ALPHA)
        ogl.glEnable(ogl.GL_TEXTURE_2D)
        ogl.glBindTexture(ogl.GL_TEXTURE_2D, self._txtr)
        ogl.glTexEnvf(ogl.GL_TEXTURE_ENV, ogl.GL_TEXTURE_ENV_MODE,
                      ogl.GL_REPLACE)
        ogl.glTexParameterfv(ogl.GL_TEXTURE_2D, ogl.GL_TEXTURE_MIN_FILTER,
                           ogl.GL_LINEAR)
        ogl.glBegin(ogl.GL_QUADS)
        ogl.glTexCoord2f(0.0, 1.0)
        ogl.glVertex3f(*self.dims[0])
        ogl.glTexCoord2f(1.0, 1.0)
        ogl.glVertex3f(*self.dims[1])
        ogl.glTexCoord2f(1.0, 0.0)
        ogl.glVertex3f(*self.dims[2])
        ogl.glTexCoord2f(0.0, 0.0)
        ogl.glVertex3f(*self.dims[3])
        ogl.glEnd()
        ogl.glDisable(ogl.GL_BLEND)
        ogl.glDisable(ogl.GL_TEXTURE_2D)
    # End of code based on Lamina module


class Visual(Stimulus):
    """A class implementing a general visual stimulus.

    All other visual stimuli should be subclassed from this class since it
    entails code for converting Pygame surfaces into OpenGL textures. This
    allows for having hardware acceleration (including waiting for the
    vertical retrace) while still being able to manipulate stimuli in an
    easy way (based on Pygame surfaces).

    """


    # The following code is based on part of the Lamina module by David Keeney
    # (https://pypi.org/project/Lamina/) with some modifications to fit it
    # into expyriment (e.g. positioning)
    @staticmethod
    def _load_texture(surf):
        """Load surface into texture object.

        Returns a texture object.

        Parameters
        ----------
        surf : pygame.Surface or numpy.array object
            surface to make texture from

        """

        txtr = ogl.glGenTextures(1)
        if isinstance(surf, pygame.Surface):
            textureData = pygame.image.tostring(surf, "RGBA", 1)
            colours = ogl.GL_RGBA
            width, height = surf.get_size()
        else:
            textureData = surf
            if textureData.shape[2] == 3:
                colours = ogl.GL_RGB
            elif textureData.shape[2] == 4:
                colours = ogl.GL_RGBA
            width, height = len(surf[0]), len(surf)
        ogl.glEnable(ogl.GL_TEXTURE_2D)
        ogl.glBindTexture(ogl.GL_TEXTURE_2D, txtr)
        ogl.glTexImage2D(ogl.GL_TEXTURE_2D, 0, colours, width, height, 0,
          colours, ogl.GL_UNSIGNED_BYTE, textureData)
        ogl.glTexParameterf(ogl.GL_TEXTURE_2D,
                            ogl.GL_TEXTURE_MAG_FILTER,
                            ogl.GL_NEAREST)
        ogl.glTexParameterf(ogl.GL_TEXTURE_2D,
                            ogl.GL_TEXTURE_MIN_FILTER,
                            ogl.GL_NEAREST)
        ogl.glDisable(ogl.GL_TEXTURE_2D)
        return txtr
    # End of code based on Lamina module

    def __init__(self, position=None, log_comment=None):
        """Create a visual stimulus.

        Parameters
        ----------
        position : (int,int), optional
            position of the stimulus
        log_comment : str, optional
            comment for the event log file

        """

        Stimulus.__init__(self, log_comment)
        if position:
            self._position = list(position)
        else:
            self._position = list(defaults.visual_position)
        self._surface = None
        self._is_preloaded = False
        self._parent = None
        self._ogl_screen = None
        self._is_compressed = False
        self._compression_filename = None

        self._was_compressed_before_preload = None

        if not _internals.active_exp.is_initialized:
            warn_message = "Stimulus created before initializing " + \
                           "(experiment defaults won't apply)!"
            print("Warning: " + warn_message)

    _exception_message = "Cannot call {0} on preloaded " \
                                     "or compressed stimulus!"

    def __del__(self):
        """ Clear surface and ogl_screen when when the objects is deconstructed.

        """

        try:
            self.clear_surface()
        except:
            pass
        if self._compression_filename is not None:
            try:
                os.remove(self._compression_filename)
            except:
                pass

    @property
    def position(self):
        """Getter for position."""

        return self._position

    @position.setter
    def position(self, value):
        """Setter for position.

        When using OpenGL, this can take longer then 1ms!

        """

        self.reposition(value)

    @property
    def polar_position(self):
        """Getter for the position in polar coordinates (radial, angle[degrees])"""

        return geometry.cartesian2polar(self._position)

    @polar_position.setter
    def polar_position(self, value):
        """Setter for the position in polar coordinates (radial, angle[degrees])

        When using OpenGL, this can take longer then 1ms!

        """

        pos = geometry.polar2cartesian(value)
        self.reposition((int(misc.round(pos[0])), int(misc.round(pos[1]))))

    @property
    def absolute_position(self):
        """Getter for absolute_position.

        Notes
        -----
        The absolute position differs for instance from the (relative) position, if the
        stimulus is plotted ontop of another stimulus, which has not the position (0,0).

        """

        if self._parent:
            return (self._parent.absolute_position[0] + self.position[0],
                    self._parent.absolute_position[1] + self.position[1])
        else:
            return self.position

    @property
    def is_compressed(self):
        """Getter for is_compressed."""

        return self._is_compressed

    @property
    def has_surface(self):
        """Getter for has_surface."""

        if self._surface is not None:
            return True
        else:
            return self.is_compressed

    @property
    def surface_size(self):
        """ Getter for surface_size."""

        return self._get_surface().get_size()

    def get_surface_copy(self):
        """Returns a copy of the Pygame surface of the stimulus

        Returns
        -------
        surface: Pygame.surface

        Notes
        -----
        see also set_surface

        """

        return self._get_surface().copy()

    def get_pixel_array(self):
        """Return a 2D array referencing the surface pixel data.

        Returns
        -------
        pixel_array: Pygame.PixelArray
            a 2D array referencing the surface pixel data

        Notes
        -----
        see also set_surface

        """

        return pygame.PixelArray(self.get_surface_copy())

    def get_surface_array(self, replace_transparent_with_colour=None):
        """Get a 3D array containing the surface pixel data.

        Returns
        -------
        surface_array : numpy.ndarray
            a 3D array containing the surface pixel data
            using RGBA coding.

        """

        if np is None:
            message = """get_rgba_array can not be used.
        The Python package 'Numpy' is not installed."""
            raise ImportError(message)

        surface = self.get_surface_copy()
        s = surface.get_size()
        pixel = pygame.surfarray.pixels3d(surface)
        alpha = pygame.surfarray.pixels_alpha(surface)
        rtn = np.empty((s[0], s[1], 4), dtype=np.int)
        if replace_transparent_with_colour is None:
            rtn[:, :, 0:3] = pixel
            rtn[:, :, 3] = alpha
        else:
            alpha = alpha.T / 255.0
            background = (np.ones(pixel.shape) * replace_transparent_with_colour)
            tmp = pixel.T * alpha + background.T * (1-alpha)
            rtn[:, :, 0:3] = tmp.T
            rtn[:, :, 3] = 255
        return rtn


    def set_surface(self, surface):
        """Set the surface of the stimulus

        This method overwrites the surface of the stimulus. It can also handle
        surfaces in form of pygame.PixelArray or Numpy 3D array (RGB or RGBA)
        representations.

        Parameters
        ----------
        surface: pygame.Surface or pygame.PixelArray or numpy.ndarray
            a representation of the new surface

        Returns
        -------
        succeeded: boolean
            setting surface was successful or not

        Notes
        -----
        CAUTION: This is an expert's method.
        The method can be used together with get_surface() & get_pixel_array()
        to apply low-level Pygame operations on stimuli. However, users should
        be aware of what they are doing, because the incorrect usage of this
        methods might affect the stability of the experiment.

        """

        if isinstance(surface, pygame.PixelArray):
            return self._set_surface(surface.make_surface())
        elif np is not None and isinstance(surface, np.ndarray):
            if surface.shape[2] == 3: # RGB
                return self._set_surface(pygame.surfarray.make_surface(surface))

            elif surface.shape[2] == 4: # RGBA
                size = surface.shape[0:2]
                new_surface = pygame.Surface(size)
                new_surface = new_surface.convert_alpha()
                [new_surface.set_at((x, y), surface[x,y,:]) \
                                for x,y in itertools.product(range(size[0]), range(size[1])) ]
                return self._set_surface(new_surface)

            else:
                return False
        elif isinstance(surface, pygame.Surface):
            return self._set_surface(surface)
        else:
            return False

    def _create_surface(self):
        """Get the surface of the stimulus.

        This method has to be overwritten for all subclasses individually!

        """

        surface = pygame.surface.Surface((0, 0))
        return surface

    def _set_surface(self, surface):
        """Set the surface (from internal use only).

        Parameters
        ----------
        surface : pygame surface
            surface to be set

        """

        if self.is_compressed:
            return False
        else:
            self._surface = surface
            return True

    def _get_surface(self):
        """Get the surface."""

        if self._surface:
            return self._surface
        else:
            if self.is_compressed:
                tmp = pygame.image.load(
                    self._compression_filename).convert_alpha()
            else:
                tmp = self._create_surface()
            return tmp

    def copy(self):
        """Deep copy of the visual stimulus.

        Returns
        -------
        copy : deep copy of self

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        if self.has_surface:
            surface_backup = self._get_surface()
            surface_copy = self._get_surface().copy()
            self._surface = None
        rtn = Stimulus.copy(self)
        if self.has_surface:
            self._surface = surface_backup
            rtn._surface = surface_copy
            rtn._is_preloaded = False
            rtn._ogl_screen = None
            rtn._is_compressed = False
            rtn._compression_filename = None
        if self.is_preloaded:
            if _internals.active_exp.screen.open_gl:
                self._ogl_screen = _LaminaPanelSurface(
                    self._get_surface(),
                    position=self.position)
            rtn.preload()
        if self.is_compressed:
            rtn.compress()
        rtn._was_compressed_before_preload = \
                self._was_compressed_before_preload
        return rtn

    def distance(self, other):
        """Surface center distance.

        This method computes the distance between the surface center
        of this and another visual stimulus.

        Parameters
        ----------
        other : stimulus
            the other visual stimulus

        Returns
        -------
        dist : float
            distance between surface centers

        """

        return geometry.XYPoint(
            self.position).distance(geometry.XYPoint(other.position))

    def replace(self, dummy):
        raise DeprecationWarning("Replace is an obsolete method. Please use reposition!")

    def reposition(self, new_position):
        """Move stimulus to a new position.

        When using OpenGL, this can take longer then 1ms!

        Parameters
        ----------
        new_position : tuple (x,y)
            translation along x and y axis

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        --------
        see also move

        """
        return self.move((new_position[0] - self.position[0],
                          new_position[1] - self.position[1]))

    def move(self, offset):
        """Moves the stimulus in 2D space.

        When using OpenGL, this can take longer then 1ms!

        Parameters
        ----------
        offset : tuple (x,y)
            translation along x and y axis

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        --------
        see also reposition

        """

        start = get_time()
        moved = False
        if offset[0] != 0:
            self._position[0] = self._position[0] + offset[0]
            moved = True
        if offset[1] != 0:
            self._position[1] = self._position[1] + offset[1]
            moved = True
        if moved and self._ogl_screen is not None:
            self._ogl_screen.refresh_position()
        return int((get_time() - start) * 1000)

    def inside_stimulus(self, stimulus, mode="visible"):
        """Check if stimulus is inside another stimulus.

        Parameters
        ----------
        stimulus : expyriment stimulus
            the other stimulus
        mode : mode (str), optional
            "visible": based on non-transparent pixels or
            "surface": based on pixels in pygame surface
            (default = visible")

        Returns
        -------
        out : bool

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        if mode == "visible":
            screen_size = _internals.active_exp.screen.surface.get_size()
            self_size = self.surface_size
            other_size = stimulus.surface_size
            self_pos = geometry.position2coordinates(self.position,
                                                     screen_size)
            self_pos[0] -= self_size[0] // 2
            self_pos[1] -= self_size[1] // 2
            if self_size[0] % 2 == 0:
                self_pos[0] += 1
            if self_size[1] % 2 == 0:
                self_pos[1] += 1
            other_pos = geometry.position2coordinates(stimulus.position,
                                                      screen_size)
            other_pos[0] -= other_size[0] // 2
            other_pos[1] -= other_size[1] // 2
            if other_size[0] % 2 == 0:
                other_pos[0] += 1
            if other_size[1] % 2 == 0:
                other_pos[1] += 1
            offset = (-self_pos[0] + other_pos[0], -self_pos[1] + other_pos[1])
            self_mask = pygame.mask.from_surface(self._get_surface())
            other_mask = pygame.mask.from_surface(stimulus._get_surface())
            overlap = self_mask.overlap_area(other_mask, offset)
            if overlap > 0 and overlap == self_mask.count():
                return True
            else:
                return False

        elif mode == "surface":
            screen_size = _internals.active_exp.screen.surface.get_size()
            sx, sy = geometry.coordinates2position(self.absolute_position,
                                                   screen_size)
            selfrect = pygame.Rect((0, 0), self.surface_size)
            if self.surface_size[0] % 2 == 0:
                sx += 1
            if self.surface_size[1] % 2 == 0:
                sy += 1
            selfrect.center = (sx, sy)
            ox, oy = geometry.coordinates2position(stimulus.absolute_position,
                                                   screen_size)
            if stimulus.surface_size[0] % 2 == 0:
                ox += 1
            if stimulus.surface_size[1] % 2 == 0:
                oy += 1
            stimrect = pygame.Rect((0, 0), stimulus.surface_size)
            stimrect.right = stimrect.right + 1
            stimrect.bottom = stimrect.bottom + 1
            stimrect.center = (ox, oy)
            if stimrect.contains(selfrect):
                return True
            else:
                return False

    def overlapping_with_stimulus(self, stimulus, mode="visible",
                                  use_absolute_position=True):
        """Check if stimulus is overlapping with another stimulus.

        Parameters
        ----------
        stimulus : expyriment stimulus
            the other stimulus
        mode : mode (str), optional
            "visible": based on non-transparent pixels or
            "surface": based on pixels in pygame surface
            (default = visible")
        use_absolute_position : bool, optional
            use absolute_position of stimuli (default) instead of position

        Returns
        -------
        overlapping : bool
            are stimuli overlapping or not
        overlap : (int, int)
            the overlap (x, y) in pixels. If mode is 'surface', the argument
            will always be None.

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        CAUTION: Please note that if a stimulus is plotted on another smaller
        stimulus, such that it is not fully visible on screen, this method will
        still check overlapping of the full stimulus! Due to a current bug in
        Pygame, we can right now not change this.

        """

        if mode == "visible":
            screen_size = _internals.active_exp.screen.surface.get_size()
            self_size = self.surface_size
            other_size = stimulus.surface_size
            if use_absolute_position:
                self_pos = geometry.position2coordinates(
                    self.absolute_position, screen_size)
                self_pos[0] -= self_size[0] // 2
                self_pos[1] -= self_size[1] // 2
                if self_size[0] % 2 == 0:
                    self_pos[0] += 1
                if self_size[1] % 2 == 0:
                    self_pos[1] += 1

                other_pos = geometry.position2coordinates(
                    stimulus.absolute_position, screen_size)
                other_pos[0] -= other_size[0] // 2
                other_pos[1] -= other_size[1] // 2
                if other_size[0] % 2 == 0:
                    other_pos[0] += 1
                if other_size[1] % 2 == 0:
                    other_pos[1] += 1

            else:
                self_pos = geometry.position2coordinates(
                    self.position, screen_size)
                self_pos[0] -= self_size[0] // 2
                self_pos[1] -= self_size[1] // 2
                if self_size[0] % 2 == 0:
                    self_pos[0] += 1
                if self_size[1] % 2 == 0:
                    self_pos[1] += 1

                other_pos = geometry.position2coordinates(
                    stimulus.position, screen_size)
                other_pos[0] -= other_size[0] // 2
                other_pos[1] -= other_size[1] // 2
                if other_size[0] % 2 == 0:
                    other_pos[0] += 1
                if other_size[1] % 2 == 0:
                    other_pos[1] += 1

            offset = (-self_pos[0] + other_pos[0], -self_pos[1] + other_pos[1])
            self_mask = pygame.mask.from_surface(self._get_surface())
            other_mask = pygame.mask.from_surface(stimulus._get_surface())
            overlap = self_mask.overlap_area(other_mask, offset)
            if overlap > 0:
                return True, overlap
            else:
                return False, overlap

        elif mode == "surface":
            screen_size = _internals.active_exp.screen.surface.get_size()
            if use_absolute_position:
                sx, sy = geometry.coordinates2position(
                    self.absolute_position, screen_size)
                ox, oy = geometry.coordinates2position(
                    stimulus.absolute_position, screen_size)
            else:
                sx, sy = geometry.coordinates2position(
                    self.position, screen_size)
                ox, oy = geometry.coordinates2position(
                    stimulus.position, screen_size)
            selfrect = pygame.Rect((0, 0), self.surface_size)
            if self.surface_size[0] % 2 == 0:
                sx += 1
            if self.surface_size[1] % 2 == 0:
                sy += 1
            selfrect.center = (sx, sy)
            stimrect = pygame.Rect((0, 0), stimulus.surface_size)
            if stimulus.surface_size[0] % 2 == 0:
                ox += 1
            if stimulus.surface_size[1] % 2 == 0:
                oy += 1
            stimrect.right = stimrect.right + 1
            stimrect.bottom = stimrect.bottom + 1
            stimrect.center = (ox, oy)
            if selfrect.colliderect(stimrect):
                return True, None
            else:
                return False, None

    def overlapping_with_position(self, position, mode="visible",
                                  use_absolute_position=True):
        """Check if stimulus is overlapping with a certain position.

        Parameters
        ----------
        position : (int, int)
            position to check for overlapping
        mode : mode (str), optional
            "visible": based on non-transparent pixels or
            "rectangle": based on pixels in pygame surface
            (default = visible")
        use_absolute_position : bool, optional
            use absolute_position of stimulus (default) instead of position

        Returns
        -------
        overlapping : bool

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        CAUTION: Please note that if a stimulus is plotted on another smaller
        stimulus, such that it is not fully visible on screen, this method will
        still check overlapping of the full stimulus! Due to a current bug in
        Pygame, we can right now not change this.

        """

        if mode == "visible":
            screen_size = _internals.active_exp.screen.surface.get_size()
            self_size = self.surface_size
            if use_absolute_position:
                self_pos = geometry.position2coordinates(
                    self.absolute_position, screen_size)
                self_pos[0] -= self_size[0] // 2
                self_pos[1] -= self_size[1] // 2
                if self_size[0] % 2 == 0:
                    self_pos[0] += 1
                if self_size[1] % 2 == 0:
                    self_pos[1] += 1

            else:
                self_pos = geometry.position2coordinates(
                    self.position, screen_size)
                self_pos[0] -= self_size[0] // 2
                self_pos[1] -= self_size[1] // 2
                if self_size[0] % 2 == 0:
                    self_pos[0] += 1
                if self_size[1] % 2 == 0:
                    self_pos[1] += 1

            pos = geometry.position2coordinates(position, screen_size)
            offset = (int(pos[0] - self_pos[0]), int(pos[1] - self_pos[1]))
            self_mask = pygame.mask.from_surface(self._get_surface())
            overlap = False
            if 0 <= offset[0] < self_size[0] and 0 <= offset[1] < self_size[1]:
                overlap = self_mask.get_at(offset)
                if overlap > 0:
                    overlap = True
                else:
                    overlap = False
            return overlap

        elif mode == "surface":
            screen_size = _internals.active_exp.screen.surface.get_size()
            if use_absolute_position:
                sx, sy = geometry.coordinates2position(self.absolute_position,
                                                       screen_size)
            else:
                sx, sy = geometry.coordinates2position(self.position,
                                                       screen_size)
            selfrect = pygame.Rect((0, 0), self.surface_size)
            if self.surface_size[0] % 2 == 0:
                sx += 1
            if self.surface_size[1] % 2 == 0:
                sy += 1
            selfrect.center = (sx, sy)
            p = geometry.coordinates2position(position, screen_size)
            p = (position[0] + screen_size[0] // 2,
                 position[1] + screen_size[1] // 2)
            p = geometry.coordinates2position(position, screen_size)
            if selfrect.collidepoint(p):
                return True
            else:
                return False

    def plot(self, stimulus):
        """Plot the stimulus on the surface of another stimulus.

        Use this to plot more than one stimulus and to present them at the
        same time afterwards by presenting the stimulus on which they were
        plotted on.

        Parameters
        ----------
        stimulus : expyriment stimulus
            stimulus to whose surface should be plotted

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not stimulus._set_surface(stimulus._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "plot()"))
        stimulus.unload(keep_surface=True)
        self._parent = stimulus
        surface = self._get_surface()
        surface_size = surface.get_size()
        rect = pygame.Rect((0, 0), surface_size)
        x, y = geometry.position2coordinates(self.position,
                                             stimulus.surface_size)
        if surface_size[0] % 2 == 0:
            x += 1
        if surface_size[1] % 2 == 0:
            y += 1
        rect.center = (x, y)
        stimulus._get_surface().blit(surface, rect)
        if self._logging:
            _internals.active_exp._event_file_log(
                "Stimulus,plotted,{0},{1}".format(self.id, stimulus.id), 2)
        return int((get_time() - start) * 1000)

    def clear_surface(self):
        """Clear the stimulus surface.

        Surfaces are automatically created after any surface operation
        (presenting, plotting, rotating, scaling, flipping etc.) and preloading.
        If the stimulus was preloaded, this method unloads the stimulus.
        This method is functionally equivalent with unload(keep_surface=False).

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if self.is_preloaded:
            self.unload(keep_surface=False)
        self._is_compressed = False
        self._set_surface(None)
        if self._logging:
            _internals.active_exp._event_file_log(
                            "Stimulus,surface cleared,{0}".format(self.id), 2)
        return int((get_time() - start) * 1000)

    def compress(self):
        """"Compress the stimulus.

        This will create a temporary file on the disk where the surface of the
        stimululs is written to.
        The surface will now be read from the disk to free memory.
        Compressed stimuli cannot do surface operations!
        Preloading comressed stimuli is possible and highly recommended.
        Depending on the size of the stimulus, this method may take some time
        to compute!

        Returns
        -------
        time : int
            the time it took to execute this method

        """

        start = get_time()
        if self.is_compressed is False:
            if self._compression_filename is None:
                fid, self._compression_filename = tempfile.mkstemp(
                    dir=defaults.tempdir, suffix=".tga")
                os.close(fid)
            pygame.image.save(self._get_surface(), self._compression_filename)
            self._is_compressed = True
            self._surface = None

            if self._logging:
                _internals.active_exp._event_file_log(
                                "Stimulus,compressed,{0}".format(self.id), 2)
        return int((get_time() - start) * 1000)

    def decompress(self):
        """Decompress the stimulus.

        This will decompress the stimulus.
        The surface will now be read from memory again.
        Depending on the size of the stimulus, this method may take some time
        to compute!

        Returns
        -------
        time : int
            the time it took to execute this method

        """

        start = get_time()
        if self.is_compressed:
            self._surface = pygame.image.load(
                self._compression_filename).convert_alpha()
            self._is_compressed = False

            if self._logging:
                _internals.active_exp._event_file_log(
                            "Stimulus,decompressed,{0}".format(self.id), 2)
        return int((get_time() - start) * 1000)

    def preload(self, inhibit_ogl_compress=False):
        """Preload the stimulus to memory.

        This will prepare the stimulus for a fast presentation.
        In OpenGL mode this method creates an OpenGL texture based on the
        surface of the stimulus.
        When OpenGL is switched off, this method will create a surface if it
        doesn't exists yet.
        If stimuli are not preloaded manually, this will happen
        automatically during presentation. However, stimulus presentation will
        take some time then!

        Always preload your stimuli when a timing accurate presentation is
        needed!

        Parameters
        ----------
        inhibit_ogl_compress : bool, optional
            inhibits OpenGL stimuli to be automatically compressed
            (default=False)

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not _internals.active_exp.is_initialized:
            message = "Can't preload stimulus. Expyriment needs to be " + \
                      "initialized before preloading a stimulus."
            raise RuntimeError(message)
        self._was_compressed_before_preload = self.is_compressed
        if not self.is_preloaded:
            if _internals.active_exp.screen.open_gl:
                self._ogl_screen = _LaminaPanelSurface(
                    self._get_surface(),
                    position=self.position)
                if not inhibit_ogl_compress:
                    self.compress()
            else:
                self.decompress()
                self._set_surface(self._get_surface())
            self._is_preloaded = True
        if self._logging:
            _internals.active_exp._event_file_log(
                                "Stimulus,preloaded,{0}".format(self.id), 2)

        return int((get_time() - start) * 1000)

    def unload(self, keep_surface=False):
        """Unload the stimulus from memory.

        This will unload preloaded stimuli.
        In OpenGL mode, this method will remove the reference to the OpenGL
        texture and the surface (when 'keep_surface' is False).
        When OpenGL is switched off, the reference to the surface will be
        removed (when 'keep_surface' is False).

        Parameters
        ----------
        keep_surface : bool, optional
            keep the surface after unload (default=False)

        Returns
        -------
        time : int
            the time it took to execute this method

        See Also
        --------
        clear_surface

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if _internals.active_exp.screen.open_gl:
            self._ogl_screen = None
            if self.is_preloaded and not self._was_compressed_before_preload \
                and keep_surface:
                self.decompress()
        else:  # Pygame surface
            if self.is_preloaded and self._was_compressed_before_preload \
                and keep_surface:
                self.compress()
        if self.is_preloaded and self._logging:
            _internals.active_exp._event_file_log("Stimulus,unloaded,{0}"\
                                       .format(self.id), 2)
        if not keep_surface:
            self._is_compressed = False
            self._surface = None
            if self._logging:
                _internals.active_exp._event_file_log("Stimulus,surface cleared,{0}"\
                                       .format(self.id), 2)

        self._is_preloaded = False
        return int((get_time() - start) * 1000)

    @property
    def is_preloaded(self):
        """Getter for is_preloaded."""

        return self._is_preloaded

    def present(self, clear=True, update=True, log_event_tag=None):
        """Present the stimulus on the screen.

        This clears and updates the screen automatically.
        When not preloaded, depending on the size of the stimulus, this method
        can take some time to compute!

        Parameters
        ----------
        clear : bool, optional
            if True the screen will be cleared automatically
            (default = True)
        update : bool, optional
            if False the screen will be not be updated automatically
            (default = True)
        log_event_tag : numeral or string, optional
            if log_event_tag is defined and if logging is switched on for this
            stimulus (default), a summary of the inter-event-intervalls are
            appended at the end of the event file

        Returns
        -------
        time : int
            the time it took to execute this method

        """

        if not _internals.active_exp.is_initialized or\
                             _internals.active_exp.screen is None:
            raise RuntimeError("Cannot not find a screen!")

        start = get_time()
        preloading_required = not(self.is_preloaded)

        if clear:
            _internals.active_exp.screen.clear()
        if preloading_required:
            # Check if stimulus has surface
            keep_surface = self.has_surface
            self.preload(inhibit_ogl_compress=True)

        if _internals.active_exp.screen.open_gl:
            self._ogl_screen.display()
        else:
            screen = _internals.active_exp.screen.surface
            rect = pygame.Rect((0, 0), self.surface_size)
            screen_size = screen.get_size()
            x, y = geometry.position2coordinates(self.position, screen_size)
            if self.surface_size[0] % 2 == 0:
                x += 1
            if self.surface_size[1] % 2 == 0:
                y += 1
            rect.center = (x, y)
            screen.blit(self._get_surface(), rect)
        if self._logging:
            _internals.active_exp._event_file_log("Stimulus,drawn,{0}"\
                                   .format(self.id), 2,
                                 log_event_tag=log_event_tag)
        if update:
            _internals.active_exp.screen.update()
        if self._logging:
            _internals.active_exp._event_file_log("Stimulus,presented,{0}"\
                                   .format(self.id), 1,
                                 log_event_tag=log_event_tag)
        if preloading_required:
            self.unload(keep_surface=keep_surface)

        return int((get_time() - start) * 1000)

    def save(self, filename):
        """Save the stimulus as image.

        Parameters
        ----------
        filename : str
            name of the file to write (possible extensions are BMP, TGA, PNG,
            or JPEG with TGA being the default)

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        parts = filename.split(".")
        if len(parts) > 1:
            parts[-1] = parts[-1].lower()
        else:
            parts.append("tga")
        filename = ".".join(parts)
        pygame.image.save(self._get_surface(), unicode2byte(filename))

    def picture(self):
        """Return the stimulus as Picture stimulus.

        This will create a temporary file on the hard disk where the image is
        saved to.

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        from . import _picture
        fid, location = tempfile.mkstemp(dir=defaults.tempdir,
                                         suffix=".tga")
        os.close(fid)
        pygame.image.save(self._get_surface(), location)
        return _picture.Picture(filename=location)

    def rotate(self, degree, filter=True):
        """Rotate the stimulus.

        This is a surface operation. After this, a surface will be present!
        Rotating goes along with a quality loss. Thus, rotating an already
        rotated stimulus is not a good idea.

        Parameters
        ----------
        degree : int
            degree to rotate counterclockwise
        filter : bool, optional
            filter the surface content for better quality (default = True)

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "rotate()"))
        self.unload(keep_surface=True)
        if filter:
            self._set_surface(pygame.transform.rotozoom(self._get_surface(),
                                                        degree, 1.0))
        else:
            self._set_surface(pygame.transform.rotate(self._get_surface(),
                                                      degree))
        if self._logging:
            _internals.active_exp._event_file_log(
                "Stimulus,rotated,{0}, degree={1}".format(self.id, degree))
        return int((get_time() - start) * 1000)

    def scale(self, factors):
        """Scale the stimulus.

        This is a surface operation. After this, a surface will be present!
        Negative scaling values will flip the stimulus.
        Scaling goes along with a quality loss. Thus, scaling an already
        scaled stimulus is not a good idea.

        Parameters
        ----------
        factors : (int, int) or (float, float)
            tuple representing the x and y factors to scale or a single number.
            In the case of a single number x and y scaling will be the
            identical (i.e., proportional scaling)

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "scale()"))
        self.unload(keep_surface=True)
        flip = [False, False]
        if isinstance(factors, (int, float)):
            factors = [factors, factors]
        else:
            factors = list(factors)
        if factors[0] < 0:
            flip[0] = True
            factors[0] = abs(factors[0])
        if factors[1] < 0:
            flip[1] = True
            factors[1] = abs(factors[1])
        self._set_surface(pygame.transform.smoothscale(
            self._get_surface(),
            (int(misc.round(self.surface_size[0] * factors[0])),
             int(misc.round(self.surface_size[1] * factors[1])))))
        if True in flip:
            self.flip(flip)
        if self._logging:
            _internals.active_exp._event_file_log(
                "Stimulus,scaled,{0}, factors={1}".format(self.id, factors), 2)
        return int((get_time() - start) * 1000)

    def scale_to_fullscreen(self, keep_aspect_ratio=True):
        """Scale the stimulus to fullscreen.

        This is a surface operation. After this, a surface will be present!
        Scaling goes along with a quality loss. Thus, scaling an already
        scaled stimulus is not a good idea.

        Parameters
        ----------
        keep_aspect_ratio : boolean, optional
            if this boolean is False, stimulus will be stretched so that it
            fills out the whole screen (default = False)

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "scale_to_fullscreen()"))
        surface_size = self.surface_size
        screen_size = _internals.active_exp.screen.surface.get_size()
        scale = (screen_size[0]/float(surface_size[0]),
                     screen_size[1]/float(surface_size[1]))
        if keep_aspect_ratio:
            scale = [min(scale)]*2
        self.scale(factors=scale)
        return int((get_time() - start) * 1000)

    def flip(self, booleans):
        """Flip the stimulus.

        This is a surface operation. After this, a surface will be present!

        Parameters
        ----------
        booleans : (bool, bool)
            booleans to flip or not

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "flip()"))
        self.unload(keep_surface=True)
        self._set_surface(pygame.transform.flip(self._get_surface(),
                                                  booleans[0], booleans[1]))
        if self._logging:
            _internals.active_exp._event_file_log(
            "Stimulus,flipped,{0}, booleans={1}".format(self.id, booleans), 2)
        return int((get_time() - start) * 1000)

    def blur(self, level):
        """Blur the stimulus.

        This blurs the stimulus, by scaling it down and up by the factor of
        'level'.

        Parameters
        ----------
        level : int
            level of bluring

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        self.scale((1.0 / level, 1.0 / level))
        self.scale((level, level))
        if self._logging:
            _internals.active_exp._event_file_log(
                "Stimulus,blured,{0}, level={1}".format(self.id, level), 2)
        return int((get_time() - start) * 1000)

    def scramble(self, grain_size):
        """Scramble the stimulus.

        Attention: If the surface size is not a multiple of the grain size,
        you may loose some pixels on the edge.

        Parameters
        ----------
        grain_size : int or (int, int)
            size of a grain (use tuple of integers for different width & height)

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """

        start = get_time()
        if isinstance(grain_size, int):
            grain_size = [grain_size, grain_size]
        # Make Rect list
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "scramble()"))
        s = self.surface_size
        source = []
        for r in range(s[1] // int(grain_size[1])):
            for c in range(s[0] // int(grain_size[0])):
                xy = (c * int(grain_size[0]), r * int(grain_size[1]))
                source.append(pygame.Rect(xy, grain_size))
        # Make copy and shuffle
        dest = copy.deepcopy(source)
        random.shuffle(dest)
        # Create a new surface
        tmp_surface = pygame.surface.Surface(
            s, pygame.SRCALPHA).convert_alpha()
        for n, s in enumerate(source):
            tmp_surface.blit(self._get_surface(), dest[n], s)
        self._set_surface(tmp_surface)

        if self._logging:
            _internals.active_exp._event_file_log(
                            "Stimulus,scrambled,{0}, grain_size={1}".format(
                                     self.id, grain_size), 2)
        return int((get_time() - start) * 1000)

    def add_noise(self, grain_size, percentage, colour):
        """Add visual noise on top of the stimulus.

        This function might take very long for large stimuli.

        Parameters
        ----------
        grain_size : int
            size of the grains for the noise
        percentage : int
            percentage of covered area
        colour : (int, int, int)
            colour (RGB) of the noise

        Returns
        -------
        time : int
            the time it took to execute this method

        Notes
        -----
        Depending on the size of the stimulus, this method may take some time
        to compute!

        """
        from . import _rectangle
        start = get_time()
        if not self._set_surface(self._get_surface()):
            raise RuntimeError(Visual._exception_message.format(
                "add_noise()"))
        self.unload(keep_surface=True)
        number_of_pixel_x = int(self.surface_size[0] // grain_size) + 1
        number_of_pixel_y = int(self.surface_size[1] // grain_size) + 1
        seq = list(range(number_of_pixel_x * number_of_pixel_y))
        random.seed()
        random.shuffle(seq)

        for idx in seq[:int(len(seq) * (percentage) / 100.0)]:
            x = (idx % number_of_pixel_x) * grain_size
            x = int(self.surface_size[0] // 2 - grain_size // 2 - x)
            y = (idx // number_of_pixel_x) * grain_size
            y = int(self.surface_size[1] // 2 - grain_size // 2 - y)
            dot = _rectangle.Rectangle(size=(grain_size, grain_size),
                            position=(x, y), colour=colour)
            dot.plot(self)
        if self._logging:
            _internals.active_exp._event_file_log(
                    "Stimulus,noise added,{0}, grain_size={1}, percentage={2}"\
                        .format(self.id, grain_size, percentage))
        return int((get_time() - start) * 1000)



"""Data Preprocessing Module.

This module contains several classes and functions that help
to handle, preprocessing and aggregate Expyriment data files.

"""

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._data_preprocessing import read_datafile, get_experiment_duration
from ._data_preprocessing import write_csv_file, write_concatenated_data
from ._data_preprocessing import Aggregator



#!/usr/bin/env python

"""
Create the API reference HTML documentation.

This script creates an HTML API documentation, based on the docstrings.
Importantly, it follows the actual namespace hierarchy and ignores everything
that starts with _.

"""


from __future__ import print_function

import __future__
import builtins


__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = '21-11-2011'


import os
import sys
import inspect  
import types
import imp


exclude = dir(builtins) + dir(__future__) + ['builtins', 'ModuleType',
                                             'True', 'False']

def inspect_members(item):
    members = inspect.getmembers(eval(item))
    modules = []
    classes = []
    methods = []
    functions = []
    attributes = []
    for member in members:
        if member[0][0:1] != '_' and member[0] not in exclude:
            #print(member)
            if inspect.ismodule(member[1]):
                modules.append(member)
            elif inspect.isclass(member[1]):
                classes.append(member)
            elif inspect.isfunction(member[1]):
                functions.append(member)
            elif inspect.ismethod(member[1]):
                methods.append(member)
            else:
                attributes.append(member)

    return modules, classes, methods, functions, attributes


def format_doc(doc):
    lines = doc.split("\n")
    cutoff = 0
    if len(lines) > 2:
        for char in lines[2]:
            if char != " ":
                break
            else:
                cutoff += 1
    for counter, line in enumerate(lines):
        if cutoff > 0 and line.startswith(" "*cutoff):
            lines[counter] = "    " + line[cutoff:]
        else:
            lines[counter] = "    " + line
    doc = "\n".join(lines)
    if doc[-1] != " ":
        doc = doc + "\n\n"
    return doc

def create_module_section(modules, item):
    section = ""
    for m in modules:
        section = section + "<span class='separator'>[</span><span class='module_name'><a href='" +\
        item + "." + m[0] + ".html'>" + m[0] + "</a></span><span class='separator'>] </span>"
    if section != "":
        section = "</pre><code><span class='section_heading'>Modules</span><br /><br />" + section + "<br /><br /><br /><br /></code><pre>"
    return section

def create_classes_section(classes, item):
    section = ""
    for c in classes:
        section = section + "<span class='separator'>[</span><span class='class_name'><a href='#" + c[0] + "'>" + c[0] + "</a></span><span class='separator'>] </span>"
    if section != "":
        section = "</pre><code><span class='section_heading'>Classes</span><br /><br />" + section + "<br /><br /><br /><br /></code><pre>"
    return section

def create_classes_details_section(classes, item):
    section = ""
    for c in classes:
        print(c)
        definition = "".join(inspect.getsourcelines(c[1])[0])
        start = definition.find("def __init__(self") + 17
        end = definition.find(")", start)
        if definition[start] == ",":
            call = "<span class='class_call'>(" + definition[start + 1:end].lstrip() + ")</span>"
        else:
            call = "<span class='class_call'>()</span>"
        call = call.replace(" " * 16, " " * len(c[0]))
        start = definition.find('"""', start) + 3
        end = definition.find('"""', start)
        doc = definition[start:end]
        if doc is None:
            doc = ""
        doc = format_doc(doc)
        section = section + "<span class='class_name'><a name='" + c[0] + "'></a><a href='" + item + "." + c[0] + ".html'>" + c[0] + "</a></span>" + call + "<span class='definition'><br /><br />" + doc + "</span><br />"
    if section != "":
        section = "<span class='section_heading'>Details (Classes)</span><br /><br />" + section + "<br />"
    return section

def create_methods_section(methods):
    section = ""
    for m in methods:
        section = section + "<span class='separator'>[</span><span class='method_name'><a href='#" + m[0] + "'>" + m[0] + "</a></span><span class='separator'>] </span>"
    if section != "":
        section = "</pre><code><span class='section_heading'>Methods</span><br /><br />" + section + "<br /><br /><br /><br /></code><pre>"
    return section

def create_methods_details_section(methods):
    section = ""
    for m in methods:
        definition = "".join(inspect.getsourcelines(m[1])[0])
        start = definition.find("(self") + 1
        end = definition.find(")")
        if definition[start + 4] == ",":
            call = "(" + definition[start + 5:end].lstrip() + ")"
        else:
            call = "()"
        call = call.replace("        " + " "*len(m[0]) + " ",
                            " "*len(m[0]) + " ")
        doc = m[1].__doc__
        if doc is None:
            doc = ""
        doc = format_doc(doc)
        section = section + "<span class='method_name'><a name='" + m[0] + "'></a>" + m[0] + "</span><span class='method_call'>" + call + "</span><br /><br />" + "<span class='definition'>" + doc + "</span><br />"
    if section != "":
        section = "<span class='section_heading'>Details (Methods)</span><br /><br />" + section + "<br />"
    return section

def create_functions_section(functions):
    section = ""
    for f in functions:
        section = section + "<span class='separator'>[</span><span class='function_name'><a href='#" + f[0] + "'>" + f[0] + "</a></span><span class='separator'>] <span/>"
    if section != "":
        section = "</pre><code><span class='section_heading'>Functions</span><br /><br />" + section + "<br /> <br /><br /><br /></code><pre>"
    return section

def create_functions_details_section(functions):
    section = ""
    for f in functions:
        definition = "".join(inspect.getsourcelines(f[1])[0])
        start = definition.find("(") + 1
        end = definition.find(")")
        call = definition[start:end]
        call = call.replace("    " + " "*len(f[0]) + " ",
                            " "*len(f[0] + " "))
        doc = f[1].__doc__
        if doc is None:
            doc = ""
        doc = format_doc(doc)
        section = section + "<span class='function_name'><a name='" + f[0] + \
             "'></a>" + f[0] + "</span><span class='function_call'>(" + call + \
             ")</span><br /><br />" + "<span class='definition'>" + doc + \
             "</span><br />"
    if section != "":
        section = "<span class='section_heading'>Details (Functions)</span>" + \
                    "<br /><br />" + section + "<br /> "
    return section

def create_attributes_section(attributes):
    section = ""
    for a in attributes:
        section = section + "<span class='separator'>[</span><span class='attribute_name'>" + \
        a[0] + "</span><span class='separator'>] </span>"
    if section != "":
         section = "</pre><code><span class='section_heading'>Attributes</span><br /><br />" + \
          section + "<br /><br /><br /><br /></code><pre>"
    return section

def create_page(item):
    modules, classes, methods, functions, attributes = inspect_members(item)
    trace = ""
    parts = item.split(".")
    if len(parts) > 1:
        for cnt, p in enumerate(parts[:-1]):
            if cnt > 0:
                link = "<a href='" + ".".join(parts[:cnt + 1]) + ".html'>" + parts[cnt]
            else:
                link = "<a href='" + p + ".html'>" + parts[cnt]
            trace = trace + link + "</a>" + "."
    title = "<span class='title'>" + trace + parts[-1] + "</span>"
    title = title + "<br />"
    doc = eval(item).__doc__
    if doc is None:
        doc = ""
    doc = format_doc(doc)
    doc = doc.lstrip()
    doc = doc.replace("\n    ", "\n")
    file_path = os.path.split(os.path.abspath(__file__))[0]
    p = os.path.abspath('{0}/../../CHANGES.md'.format(file_path))
    version_nr = "{0}"
    with open(p) as f:
        for line in f:
            if line[0:8].lower() == "upcoming":
                 version_nr += "+"
            if line[0:7] == "Version":
                version_nr = version_nr.format(line[8:13])
                break
    page = """
<html>
<head>
<link rel="shortcut icon" href="favicon.ico" type="image/x-icon" />
<title>API reference for {0} ({1})</title>
{2}
</head>
<pre>
{3}
<span class='definition'>{4}</span><br />
{5}{6}{7}{8}{9}
<hr >
{10}{11}{12}
</pre>
</html>""".format(item, version_nr, css, title, doc,
                  create_module_section(modules, item),
                  create_classes_section(classes, item),
                  create_methods_section(methods),
                  create_functions_section(functions),
                  create_attributes_section(attributes),
                  create_classes_details_section(classes, item),
                  create_methods_details_section(methods),
                  create_functions_details_section(functions))

    p = os.path.abspath("{0}/{1}.html".format(os.path.split(os.path.abspath(__file__))[0],
                        item))
    print("create", p)
    with open(p, 'w') as f:
        f.write(page)
    if modules:
        for m in modules:
            create_page(item + "." + m[0])

    if classes:
        for c in classes:
            create_page(item + "." + c[0])


if __name__ == "__main__":
    css = """
body {font-size:100%; color: #393333}
a {text-decoration:none; color:blue;}
a:visited {text-decoration:none;}
a:hover {text-decoration: underline;}
hr {color: #cccccc; background-color: #cccccc;}
.title {font-size:1.5em; font-weight:bold;}
.title_call {font-size:1.5em; font-weight:normal; font-style:italic;}
.definition {font-size:1em; color:#555555}
.section_heading {font-size:1.5em; font-weight:bold; color: #d14836;}
.separator {color:#777777;}
.module_name {font-size:1em; font-weight:bold;}
.class_name {font-size:1em; font-weight:bold;}
.class_call {font-size:1em; font-weight:normal; font-style: italic;}
.method_name {font-size:1em; font-weight:bold;}
.method_call {font-size:1em; font-weight:normal; font-style:italic;}
.function_name {font-size:1em; font-weight:bold;}
.function_call {font-size:1em; font-weight:normal; font-style:italic;}
.attribute_name {font-weight:bold;}"""

    css = "<style TYPE='text/css'>\n<!--\n" + css + "\n--!>\n</style>"

    cmd_folder = os.path.dirname(os.path.abspath(__file__+"/../../"))
    if cmd_folder not in sys.path:
        sys.path.insert(0, cmd_folder)
        import expyriment
        import expyriment.io.extras
        import expyriment.design.extras
        import expyriment.stimuli.extras
        import expyriment.misc.extras
        sys.path.remove(cmd_folder)
    create_page("expyriment")

    p = os.path.abspath("{0}/index.html".format(os.path.split(os.path.abspath(__file__))[0]))
    print("create", p)
    with open(p, 'w') as f:
        content = """<html><head><meta http-equiv="refresh" content="0; URL=expyriment.html"></head>"""
        f.write(content)



"""
The geometry module.

This module contains miscellaneous geometry functions for expyriment.

"""

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ._geometry import coordinates2position, position2coordinates, position2coordinate
from ._geometry import position2visual_angle, visual_angle2position
from ._geometry import cartesian2polar, polar2cartesian, tuples2points
from ._geometry import points_to_vertices, points2vertices, lines_intersect, XYPoint
from ._geometry import lines_intersection_point
from ._basic_shapes import vertices_triangle, vertices_rectangle, vertices_regular_polygon
from ._basic_shapes import vertices_frame, vertices_trapezoid, vertices_parallelogram, vertices_cross


"""
The design._structure module of expyriment.

This module contains a class implementing the experiment structure.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os
try:
    import locale
except ImportError:
    locale = None  # Does not exist on Android
import sys
from types import FunctionType
import codecs
import re
try:
    import csv
except ImportError:
    from ..misc import _csv_reader_android as csv
from copy import deepcopy

from . import defaults
from .. import _internals
from ..misc import constants, Clock, unicode2byte, byte2unicode, py2py3_sort_array
from .randomize import rand_int, shuffle_list
from . import permute

_FACTOR_NOT_EXIST = "The factor '{0}' does not exist!\nUse has_factor(name) to check if a factor is defined."
_BWS_FACTOR_NOT_EXIST = "The bws-factor '{0}' does not exist!\nUse has_bws_factor(name) to check if a bws-factor is defined."

class Experiment(object):
    """A class implementing a basic experiment."""

    def __init__(self, name=None, foreground_colour=None,
                 background_colour=None, text_font=None, text_size=None,
                 filename_suffix=None):
        """Create an experiment.

        Parameters
        ----------
        name : str, optional
            name of the experiment
        foreground_colour : (int, int, int), optional
        background_colour : (int,int,int), optional
        text_font : str, optional
        text_size : int, optional
        filename_suffix : str, optional
            additional suffix that will be added to the main event
            and data filename

        """
        if name is not None:
            self._name = name
        elif defaults.experiment_name is not None:
            self._name = defaults.experiment_name
        else:
            self._name = os.path.split(sys.argv[0])[1].replace(".py", "")
        if background_colour is not None:
            self._background_colour = background_colour
        else:
            self._background_colour = \
                defaults.experiment_background_colour
        if foreground_colour is not None:
            self._foreground_colour = foreground_colour
        else:
            self._foreground_colour = \
                defaults.experiment_foreground_colour
        if text_font is not None:
            self._text_font = text_font
        else:
            self._text_font = defaults.experiment_text_font
        if text_size is not None:
            self._text_size = text_size
        else:
            self._text_size = defaults.experiment_text_size
        if filename_suffix is not None:
            self._filename_suffix = filename_suffix
        else:
            self._filename_suffix = defaults.experiment_filename_suffix

        self.clear_bws_factors()
        self._data_variable_names = []
        self._experiment_info = []
        self._blocks = []
        self._block_id_counter = 0

        self._is_started = False
        self._is_initialized = False
        self._keyboard = None
        self._mouse = None
        self._clock = None
        self._subject = None
        self._screen = None
        self._data = None
        self._events = None
        self._log_level = None  # will be set from initialize
        self._wait_callback_function = None

    @property
    def name(self):
        """Getter for name."""

        return self._name

    @property
    def blocks(self):
        """Getter for blocks."""

        return self._blocks

    @property
    def background_colour(self):
        """Getter for background_colour."""

        return self._background_colour

    @property
    def foreground_colour(self):
        """Getter for foreground_colour."""

        return self._foreground_colour

    @property
    def text_font(self):
        """Getter for text_font."""

        return self._text_font

    @property
    def text_size(self):
        """Getter for text_size."""

        return self._text_size

    @property
    def filename_suffix(self):
        """Getter for filename suffix."""

        return self._filename_suffix

    @property
    def screen(self):
        """Getter for global screen."""

        return self._screen

    @property
    def data(self):
        """Getter for main data file."""

        return self._data

    @property
    def events(self):
        """Getter for main event files."""

        return self._events

    @property
    def log_level(self):
        """Getter for event logging. """

        return self._log_level

    @property
    def clock(self):
        """Getter for global clock."""

        return self._clock

    @property
    def keyboard(self):
        """Getter for global keyboard."""

        return self._keyboard

    @property
    def mouse(self):
        """Getter for global keyboard."""

        return self._mouse

    @property
    def subject(self):
        """Getter for global subject id."""

        return self._subject

    @property
    def is_started(self):
        """Getter for is_started."""

        return self._is_started

    @property
    def is_initialized(self):
        """Getter for is_initialized."""

        return self._is_initialized

    def __str__(self):
        tmp_str = u"Experiment: {0}\n".format(self.name)
        if len(self.bws_factor_names) <= 0:
            tmp_str = tmp_str + "no between subject factors\n"
        else:
            tmp_str = tmp_str + "between subject factors (permutation type: "
            if self.bws_factor_randomized:
                tmp_str = tmp_str + "random)\n"
            else:
                tmp_str = tmp_str + "latin square)\n"
            for f in self.bws_factor_names:
                _bws_factor = \
                    [x if isinstance(x, str) else
                     repr(x) for x in self._bws_factors[f]]
                tmp_str = tmp_str + u"    {0} = [{1}]\n".format(
                    f, u", ".join(_bws_factor))
        for block in self.blocks:
            tmp_str = tmp_str + u"{0}\n".format(block.summary)
        return tmp_str

    @property
    def data_variable_names(self):
        """Getter for data_variable_names."""

        return self._data_variable_names

    @data_variable_names.setter
    def data_variable_names(self, value):
        """Setter for data_variable_names."""

        self.clear_data_variable_names()
        self.add_data_variable_names(value)

    def clear_data_variable_names(self):
        """Remove all data variable names from design."""

        self._data_variable_names = []
        if self.data is not None:
            self.data.clear_variable_names()

    def add_data_variable_names(self, variable_names):
        """Add data variable names to the design.

        Parameters
        ----------
        variables : str or lst of str
            variable names

        """
        if variable_names is None:
            return
        try:
            variable_names = list(variable_names)
        except:
            variable_names = [variable_names]
        self._data_variable_names.extend(variable_names)
        if self.data is not None:
            self.data.add_variable_names(variable_names)

    @property
    def experiment_info(self):  # experiment_info can not be cleared!
        """Getter for experiment_info."""

        return self._experiment_info

    def add_experiment_info(self, text):
        """Add experiment information to the design.

        Parameters
        ----------
        text : string or list of str
            text lines to be added as experiment information

        """

        if text is None:
            return
        elif isinstance(text, str):
            text = [text]
        else:
            try:
                text = list(text)
            except:
                text = [text]

        self._experiment_info.extend(text)
        if self.data is not None:
            for elem in text:
                self.data.add_experiment_info(elem)

    @property
    def bws_factor_randomized(self):
        """Getter for bws_factor_randomized.

        Notes
        -----
        Is between subject factor randomized? (True/False).

        If True conditions will be assigned randomized
        otherwise (default) conditions will be systematically permuted across
        subjects.

        """

        return self._bws_factor_randomized

    @bws_factor_randomized.setter
    def bws_factor_randomized(self, value):
        """Setter for bws_factor_randomized."""

        self._bws_factor_randomized = value

    def add_bws_factor(self, factor_name, conditions):
        """Add a between subject factor.

        Parameters
        ----------
        factor_name : str
            the name of the between subject factor
        conditions : list
            possible conditions of the between subject factor

        Notes
        -----
        The defined between subject factor conditions will be permuted across
        the subjects when 'using get_permuted_bws_factor_condition'.
        Factors that are added first are treated as hierarchically higher
        factors while permutation.

        See Also
        --------
        get_permuted_bws_factor_condition

        """

        try:
            conditions = list(conditions)
        except:
            conditions = [conditions]
        self._bws_factors[factor_name] = conditions
        self._bws_factors_names.append(factor_name)
        self._randomized_condition_for_subject[factor_name] = {}

    def has_bws_factor(self, factor_name):
        """Checks if a factor is defined.

        Parameters
        ----------
        factor_name : str
            factor name (str)

        Returns
        -------
        bool : boolean

        """

        return factor_name in self.bws_factor_names

    def get_bws_factor(self, factor_name, return_none_if_not_defined=False):
        """Return all conditions of this between subject factor.

        Parameters
        ----------
        factor_name : str
            factor name
        return_none_if_not_defined : boolean (default: False)
            suppresses exception and returns None if factor is not defined.

        Returns
        -------
        conditions : none or array of all conditions

        """


        try:
            cond = self._bws_factors[factor_name]
        except:
            if return_none_if_not_defined:
                return None
            else:
                raise RuntimeError(_BWS_FACTOR_NOT_EXIST.format(factor_name))
        return cond

    def get_permuted_bws_factor_condition(self, factor_name, subject_id=None):
        """Get the between subject factor condition for a subject.

        The condition for the current subject will be returned, if expyriment
        is running and the function is called without a subject_id.
        If the function is called with a subject_id, the condition for this
        particular subject will be returned.

        Parameters
        ----------
        factor_name : str
        subject_id : str, optional
            (default=None)

        Returns
        -------
        cond : str
            condition for the current subject

        Notes
        -----
        see add_bws_factor

        """

        if subject_id is None:
            if self.subject is None:  # No current subject id defined
                return None
            else:  # Use current subject
                return self.get_permuted_bws_factor_condition(
                    factor_name, subject_id=self.subject)
        else:
            conditions = self.get_bws_factor(factor_name)
            if conditions is None:
                return None  # Factor not defined
            else:
                cond_idx = 0
                if self.bws_factor_randomized:
                    try:
                        cond_idx = self._randomized_condition_for_subject[
                            factor_name][subject_id]
                    except:  # If not yet randomized for this subject, do it
                        cond_idx = rand_int(
                            0, len(self._bws_factors[factor_name]) - 1)
                        self._randomized_condition_for_subject[
                            factor_name][subject_id] = cond_idx

                else:  # Permutation
                    # (n_cond_lower_fac) total number of conditions for all
                    # hierarchically lower factors
                    n_cond_lower_fac = self.n_bws_factor_conditions
                    for fac in self.bws_factor_names:
                        n_cond_lower_fac -= len(self.get_bws_factor(fac))
                        if fac is factor_name:
                            break
                    if n_cond_lower_fac <= 0:
                        n_cond_lower_fac = 1
                    cond_idx = ((subject_id - 1) // n_cond_lower_fac) % \
                        len(self.get_bws_factor(fac))
                return self._bws_factors[factor_name][cond_idx]

    def clear_bws_factors(self):
        """Remove all between subject factors from design."""

        self._bws_factors = {}

        # Can't use dict_keys, because dicts don't keep the order
        self._bws_factors_names = []

        self._randomized_condition_for_subject = {}
        self.bws_factor_randomized = False

    @property
    def bws_factor_names(self):
        """Getter for factors keys."""
        return self._bws_factors_names

    @property
    def n_bws_factor_conditions(self):
        """Getter for n_bws_factor_conditions.

        Total number of conditions in all bws_factors.

        """

        n = 0
        for fn in self.bws_factor_names:
            n += len(self.get_bws_factor(fn))
        return n

    def set_log_level(self, loglevel):
        """Set the log level of the current experiment

        Parameters
        ----------
        loglevel : int
            The log level (0, 1, 2) of the experiment.

        Notes
        -----
        There are three event logging levels:

        - O no event logging
        - 1 normal event logging (logging of all input & output events)
        - 2 intensive logging; logs much more; please use this only for
          debugging proposes

        In most cases, it should be avoided to switch of logging (loglevel=0).
        If log files become to big due to certain repetitive events, it is
        suggested to switch of the logging of individual stimuli or IO event.
        (see the method `.set_logging()` of this objects)

        The logging of events can be also changed before initialize via the
        default value `expyriment.control.defaults.event_logging`.

        """

        self._log_level = int(loglevel)

    def add_block(self, block, copies=1):
        """Add a block to the experiment.

        Parameters
        ----------
        block : design.Block
            block to add
        copies : int, optional
            number of copies to add (default = 1)

        """

        for _x in range(0, copies):
            self._blocks.append(block.copy())
            self._blocks[-1]._id = self._block_id_counter
            self._block_id_counter += 1

        _internals.active_exp._event_file_log(
            "Experiment,block added,{0},{1}".format(
                self.name, self._blocks[-1]._id), 2)

    def add_blocks_full_factorial(self, design_dict, copies=1):
        """Add blocks of all combinations of a full factorial design.

        Factors will be defined according the design_dict parameter.
        The design is specified by a dictionary. The keys of the dictionary
        indicate the factor names. Each value of the dictionary has to be
        a list of the factor levels.

        Parameters
        ----------
        design_dict : dictionary
            keys: factor names, values: lists of factor levels
        copies : int, optional
            number of copies to add (default = 1)

        See Also
        --------
        expyriment.design.Block.add_trials_full_factorial

        """

        keys = design_dict.keys()
        levels = list(map(lambda x: len(design_dict[x]), keys))
        cnt = [0]*len(design_dict)

        while cnt is not None:
            bl = Block()
            for i,k in enumerate(keys):
                bl.set_factor(k, design_dict[ k ][ cnt[i] ])
            self.add_block(bl, copies=copies)

            cnt = _get_next_permutation(values=cnt, levels=levels)



    def remove_block(self, position):
        """Remove block from experiment.

        If no position is given, the last one is removed.

        Parameters
        ----------
        position : int
            position of the block to be removed

        """

        block = self._blocks.pop(position)

        _internals.active_exp._event_file_log(
            "Experiment,block removed,{0},{1}".format(self.name,
                                                      block.id), 2)

    def clear_blocks(self):
        """Remove all blocks from experiment."""

        self._blocks = []
        self._block_id_counter = 0

        _internals.active_exp._event_file_log("Experiment,blocks cleared", 2)

    def order_blocks(self, order):
        """Order the blocks.

        Parameters
        ----------
        order : list
            list with the new order of positions

        """

        if not len(order) == len(self._blocks):
            raise ValueError("Given order has wrong number of items!")
        blocks_new = []
        for position in order:
            blocks_new.append(self._blocks[position])
        self._blocks = blocks_new

    @property
    def n_blocks(self):
        """Getter for n_blocks.

        Number of blocks.

        """

        return len(self._blocks)

    def swap_blocks(self, position1, position2):
        """Swap two blocks.

        Parameters
        ----------
        position1 : int
            position of first block
        position2 : int
            position of second block

        """

        if position1 < len(self._blocks) and position2 < len(self._blocks):
            self._blocks[position1], self._blocks[position2] = \
                self._blocks[position2], self._blocks[position1]
            return True
        else:
            return False

    def shuffle_blocks(self, max_repetitions=None, n_segments=None):
        """Shuffle all blocks.

        Parameters
        ----------
        max_repetitions : int, optional
            see documentation of `randomize.shuffle_list`, default = None
        n_segments : int, optional
            see documentation of `randomize.shuffle_list`, default = None

        Returns
        -------
        success : bool
            returns if randomization was successful and fulfilled the specified
            constrains (see max_repetitions)

        See Also
        --------
        expyriment.design.randomize.shuffle_list

        """

        return shuffle_list(self._blocks)

    def permute_blocks(self, permutation_type, factor_names=None,
                       subject_id=None):
        """Permute the blocks.

        Parameters
        ----------
        permutation_type : int (permutation type)
            type of block order permutation (permutation type);
            permutation types defined in misc.constants:
            P_BALANCED_LATIN_SQUARE, P_CYCLED_LATIN_SQUARE, and P_RANDOM
        factor_names : list (of strings), optional
            list of the factor names to be considered while permutation;
            if factor_names are not defined (None) all factors will be used.
        subject_id : int, optional
            subject number for this permutation;
            if subject_id is defined or none (default) and experiment has
            been started, the current subject number will be used

        """

        if subject_id is None:
            if self.subject is None:
                raise RuntimeError("If Expyriment is not started, \
a subject number needs to be defined for the permutation.")
            else:
                subject_id = self.subject

        if not permute.is_permutation_type(permutation_type):
                raise AttributeError("{0} is a unknown permutation \
type".format(permutation_type))
        if factor_names is None:
            factor_names = self.block_list_factor_names

        # Get the condition combinations for the specified factors:
        all_factor_combi = []
        for b in self.blocks:
            combi = []
            for f in factor_names:
                combi.append([f, b.get_factor(f, return_none_if_not_defined=True)])
            new = True
            for c in all_factor_combi:
                if c == combi:
                    new = False
            if new:  # Add only a new combination
                all_factor_combi.append(combi)

        # Get the permutation
        if permutation_type == constants.P_BALANCED_LATIN_SQUARE:
            permutation = permute.balanced_latin_square(all_factor_combi)
            idx = (subject_id - 1) % len(permutation)
            permutation = permutation[idx]
        elif permutation_type == constants.P_CYCLED_LATIN_SQUARE:
            permutation = permute.cycled_latin_square(all_factor_combi)
            idx = (subject_id - 1) % len(permutation)
            permutation = permutation[idx]
        else:
            shuffle_list(all_factor_combi)
            permutation = all_factor_combi

        tmp = self._blocks
        self._blocks = []
        for search_combi in permutation:
            # Search tmp block for this comb
            # And add all fitting blocks  (multiple addings possible)
            for b in tmp:
                combi = []
                for f in factor_names:
                    combi.append([f, b.get_factor(f, return_none_if_not_defined=True)])
                if combi == search_combi:
                    self._blocks.append(b)

    def sort_blocks(self):
        """Sort the blocks according to their indices from low to high."""

        blocks_new = []
        id_list = [x.id for x in self._blocks]
        id_list.sort()
        for id in id_list:
            position = [i for i, x in enumerate(self._blocks)
                        if x.id == id][0]
            blocks_new.append(self._blocks[position])
        self._blocks = blocks_new

    def find_block(self, id):
        """Find the position of a block, given the id.

        Parameters
        ----------
        id : int
            block id to look for

        Returns
        -------
        pos: int
         positions as a list or None if not in block list.

        """

        positions = [i for i, x in enumerate(self._blocks) if x.id == id]
        if positions:
            return positions

    @property
    def trial_factor_names(self):
        """Getter for trial_factor_nanes.

        Get all factor names defined in the trial lists of all blocks.

        """

        factors = []
        for bl in self.blocks:
            factors.extend(bl.trial_factor_names)
        return list(set(factors))

    @property
    def block_list_factor_names(self):
        """Getter for block_list_factor_names.

        Get all factor names defined in all blocks.

        """

        factors = []
        for bl in self.blocks:
            factors.extend(bl.factor_names)
        return list(set(factors))

    @property
    def design_as_text(self):
        """Getter for desing_as_text.

        Trial list as csv table.

        """

        rtn = u"#exp: {0}\n".format(self.name)
        if len(self.experiment_info) > 0:
            for txt in self.experiment_info:
                rtn += u"#xpi: {0}\n".format(txt)
        if len(self.bws_factor_names) > 0:
            for factor_name in self.bws_factor_names:
                rtn += u"#bws: {0}=".format(factor_name)
                for txt in self.get_bws_factor(factor_name):
                    rtn += u"{0},".format(txt)
                rtn = rtn[:-1] + "\n"  # delete last comma
            rtn += u"#bws-rand: {0}\n".format(int(self.bws_factor_randomized))
        if len(self.data_variable_names) > 0:
            rtn += "#dvn: "
            for txt in self.data_variable_names:
                rtn += u"{0},".format(txt)
            rtn = rtn[:-1] + "\n"

        rtn += "block_cnt,block_id"
        bl_factors = self.block_list_factor_names
        factors = self.trial_factor_names
        for f in bl_factors:
            rtn += u",block_{0}".format(f)
        rtn += ",trial_cnt,trial_id"
        for f in factors:
            rtn += u",{0}".format(f)

        for bl_cnt, bl in enumerate(self.blocks):
            for tr_cnt, tr in enumerate(bl.trials):
                rtn += u"\n{0},{1}".format(bl_cnt, bl.id)
                for f in bl_factors:
                    rtn += u",{0}".format(bl.get_factor(f, return_none_if_not_defined=True))
                rtn += u",{0},{1}".format(tr_cnt, tr.id)
                for f in factors:
                    rtn += u",{0}".format(tr.get_factor(f, return_none_if_not_defined=True))

        return rtn

    def save_design(self, filename):
        """Save the design as list of trials to a csv file.

        The function considers only the defined trial factors and not the
        added stimuli.

        Notes
        -----
        The current version of this function does not handle between_subject
        factors and data_variables.

        Parameters
        ----------
        filename : str
            name (fullpath) of the csv file (str)

        """

        with open(filename, 'wb') as f:
            try:
                locale_enc = locale.getdefaultlocale()[1]
            except:
                locale_enc = "UTF-8"
            header = "# -*- coding: {0} -*-\n".format(locale_enc)
            f.write(unicode2byte(header + self.design_as_text))

    def load_design(self, filename, encoding=None):
        """Load the design from a csv file containing list of trials.

        The function considers only the defined trial factors and not the
        added stimuli. The existing design will be deleted.

        Notes
        -----
        The current version of this function does not handle between_subject
        factors and data_variables.

        Parameters
        ----------
        filename : str
            name (fullpath) of the csv file (str)

        encoding : str, optional
            the encoding to be used when reading from the file

        """

        delimiter = ","
        self.clear_blocks()
        block_factors = {}
        trial_factors = {}
        if encoding is None:
            with open(filename, 'r') as fl:
                first_line = fl.readline()
                encoding = re.findall("coding[:=]\s*([-\w.]+)", first_line)
                if encoding == []:
                    second_line = fl.readline()
                    encoding = re.findall("coding[:=]\s*([-\w.]+)",
                                          second_line)
                    if encoding == []:
                        encoding = [None]
        else:
            encoding = [encoding]
        with codecs.open(filename, 'rb', encoding[0], errors='replace') as fl:
            for ln in fl:
                ln = byte2unicode(ln)
                if ln[0] == "#":
                    if ln.startswith("#exp:"):
                        self._name = ln[6:].strip()
                    elif ln.startswith("#xpi:"):
                        self.add_experiment_info(ln[6:].strip())
                    elif ln.startswith("#dvn:"):
                        for tmp in ln[6:].split(","):
                            self.add_data_variable_names(tmp.strip())
                    elif ln.startswith("#bws-rand:"):
                        self.bws_factor_randomized = (ln[11] == "1")
                    elif ln.startswith("#bws:"):
                        tmp = ln[6:].split("=")
                        print(tmp[1].strip().split(","))
                        self.add_bws_factor(tmp[0], tmp[1].strip().split(","))

                else:  # data line
                    if len(block_factors) < 1:
                        # read first no-comment line --> varnames
                        for col, var in enumerate(ln.split(delimiter)):
                            var = var.strip()
                            if var.startswith("block_"):
                                var = var.replace("block_", "")
                                block_factors[col] = var
                            elif var.startswith("trial_"):
                                var = var.replace("trial_", "")
                                trial_factors[col] = var
                            else:
                                trial_factors[col] = var

                        if not("cnt" in list(block_factors.values()) and
                               "id" in list(block_factors.values()) and
                               "cnt" in list(trial_factors.values()) and
                               "id" in list(trial_factors.values())):
                            message = "Can't read design file. " + \
                                "The file '{0}' ".format(filename) + \
                                "does not contain an Expyriment trial list."
                            raise IOError(message)
                    else:
                        block_cnt = None
                        trial_cnt = None
                        # read data
                        for col, val in enumerate(ln.split(delimiter)):
                            val = val.strip()
                            # try to convert to number
                            if val.find(".") >= 0:
                                try:
                                    val = float(val)
                                except:
                                    pass
                            else:
                                try:
                                    val = int(val)
                                except:
                                    pass

                            # set value to block or trial
                            if col in block_factors:
                                if block_factors[col] == "cnt":
                                    block_cnt = val
                                    while len(self.blocks) < block_cnt + 1:
                                        self.add_block(Block())
                                elif block_factors[col] == "id":
                                    self.blocks[block_cnt]._id = val
                                else:
                                    self.blocks[block_cnt].set_factor(
                                        block_factors[col], val)

                            if col in trial_factors:
                                if trial_factors[col] == "cnt":
                                    trial_cnt = val
                                    while len(self.blocks[block_cnt].trials)\
                                            < trial_cnt + 1:
                                        self.blocks[block_cnt].add_trial(
                                            Trial())
                                elif trial_factors[col] == "id":
                                    self.blocks[block_cnt].trials[trial_cnt].\
                                        _id = val
                                else:
                                    self.blocks[block_cnt].trials[trial_cnt].\
                                        set_factor(trial_factors[col], val)

    def _event_file_log(self, log_text, log_level=1, log_event_tag=None):
        # log_level 1 = default, 2 = extensive, 0 or False = off
        """ Helper function to log event in the global experiment event file"""
        if self.is_initialized and\
                self._log_level > 0 and\
                self._log_level >= log_level and \
                self.events is not None:
            self.events.log(event=log_text, log_event_tag=log_event_tag)

    def _event_file_warn(self, warning, log_level=1):
        """ Helper function to log event in the global experiment event file"""
        if self.is_initialized and\
                self._log_level > 0 and\
                self._log_level >= log_level and \
                self.events is not None:
            self.events.log(warning)

    def log_design_to_event_file(self, additional_comment=""):
        """Log the design (as comment) to the current main event file.

        If no experiment is initialized or no event file exists the function
        will not do anything. This function will be automatically called after
        an experiment has been started.

        Parameters
        ----------
        additional_comment : str, optional
            additional comment that will be logged

        See Also
        --------
        save_design

        """

        if self.is_initialized and self.events is not None:
            self.events.log(u"design,log,{0}".format(additional_comment))
            for ln in self.design_as_text.splitlines():
                self.events.write_comment(
                    u"design: {0}".format(ln).replace(":#", "-"))
            self.events.log(u"design,logged,{0}".format(
                additional_comment))

    def register_wait_callback_function(self, function):
        """Register a wait callback function.

        The registered wait callback function will be repetitively executed in
        all Expyriment wait and event loops that wait for an external input (see below).

        If the callback function returns a control.CallbackQuitEvent the calling wait
        or event loop will be force to quit. The CallbackQuitEvent object will be then
        return (as first return value in the case of multiple return values).

        The following functions will call the currently registered callback function
        (at least once!):

            - control.wait_end_audiosystem
            - misc.Clock.wait
            - misc.Clock.wait_seconds
            - misc.Clock.wait_minutes
            - io.Keyboard.wait
            - io.Keyboard.wait_char
            - io.StreamingButtonBox.wait
            - io.EventButtonBox.wait
            - io.GamePad.wait_press
            - io.TriggerInput.wait
            - io.Mouse.wait_press
            - io.SerialPort.read_line
            - io.TextInput.get
            - io.TouchScreenButtonBox.wait
            - io.extras.CedrusResponseDevice.wait
            - stimulus.Video.wait_frame
            - stimulus.Video.wait_end

        Parameters
        ----------
        function : function
            wait function (function)

        Notes
        -----
        CAUTION! If wait callback function takes longer than 1 ms to process,
        Expyriment timing will be affected!

        See Also
        --------
        expyriment.control.CallbackQuitEvent

        """

        if isinstance(function, FunctionType):
            self._wait_callback_function = function
        else:
            raise AttributeError("register_wait_callback_function requires " +
                                 "a function as parameter")

    def unregister_wait_callback_function(self):
        """Unregister wait function."""

        self._wait_callback_function = None

    def _execute_wait_callback(self):
        """Execute wait function.

        Returns the return value of the callback function or
        False if callback is not defines.

        """

        if self._wait_callback_function is not None:
            return self._wait_callback_function()
        else:
            return False

    @property
    def is_callback_registered(self):
        """This property indicates whether a callback function has been
        registered or not.

        """
        return (self._wait_callback_function is not None)


class Block(object):
    """A class implementing an experimental block."""

    _trial_cnt_variable_name = "trial_cnt"  # variable names for csv in/output
    _trial_id_variable_name = "trial_id"

    def __init__(self, name=None):
        """Create a block.

        Parameters
        ----------
        name : str, optional
            name of the block

        """

        if name is not None:
            self._name = name
        else:
            self._name = defaults.block_name

        self._factors = {}
        self._trials = []
        self._trial_id_counter = 0
        self._id = None

    @property
    def name(self):
        """Getter for name."""

        return self._name

    @property
    def id(self):
        """Getter for id."""

        return self._id

    @property
    def trials(self):
        """Getter for trials."""

        return self._trials

    def __str__(self):
        return self._get_summary(True)

    @property
    def summary(self):
        """Getter for summary."""

        return self._get_summary(False)

    def _get_summary(self, include_trial_IDs):
        """Return a summary of the trials as string."""

        if self.name is None:
            name = ""
        else:
            name = self.name
        rtn = u"""Block {0}: {1}
    block factors: {2}
    n trials: {3}""".format(self.id, name,
                            self.factors_as_text,
                            len(self.trials))

        if include_trial_IDs:
            rtn = rtn + u"""
    trial IDs = {0}""".format([t.id for t in self.trials])
        rtn = rtn + u"""
    trial factors: """
        for f in self.trial_factor_names:
            val = []
            for tf in self.get_trial_factor_values(f):
                if tf not in val:
                    val.append(tf)
            py2py3_sort_array(val)
            val = [repr(x) if not isinstance(x, str) \
                   else x for x in val]
            rtn = rtn + u"{0} = [{1}]\n                   ".format(
                f, ", ".join(val))

        return rtn

    @property
    def factors_as_text(self):
        """Getter for factors_as_text.

        Return all factor names and values as string line.

        """

        all_factors = ""
        for f in self.factor_names:
            all_factors = all_factors + \
                u"{0} = {1}\n                   ".format(
                    f, self.get_factor(f, return_none_if_not_defined=True))
        all_factors = all_factors.rstrip()
        if len(all_factors) >= 1 and all_factors[-1] == ",":
            all_factors = all_factors[:-1]
        return all_factors

    def set_factor(self, name, value):
        """Set a factor for the block.

        Parameters
        ----------
        name : str
            factor name
        value : str or numeric
            factor value

        """

        if isinstance(value, (bytes, str, int, float)):
            self._factors[name] = value
        else:
            message = "Factor values or factor conditions must to be a " + \
                "String or a Number (i.e. float or integer).\n " + \
                "{0} is not allowed.".format(type(value))
            raise TypeError(message)

    def has_factor(self, name):
        """Checks if a factor is defined.

        Parameters
        ----------
        name : str
            factor name (str)

        Returns
        -------
        bool : boolean

        """

        return name in self.factor_names

    def get_factor(self, name, return_none_if_not_defined=False):
        """Get a factor of the block.

        Parameters
        ----------
        name : str
            factor name
        return_none_if_not_defined : boolean (default: False)
            suppresses exception and returns None if factor is not defined.

        Returns
        -------
        factor_val : str or numeric

        """

        try:
            rtn = self._factors[name]
        except:
            if return_none_if_not_defined:
                return None
            else:
                raise RuntimeError(_FACTOR_NOT_EXIST.format(name))
        return rtn

    @property
    def factor_dict(self):
        """The dictionary with all factors of the block."""

        return self._factors

    def clear_factors(self):
        """Clear all factors."""

        self._factors = {}

    @property
    def factor_names(self):
        """Getter for factor_names.

        Factor keys.

        """

        return list(self._factors.keys())

    def compare(self, block):
        """Compares this block with another block and returns `True` if all
        factors associated with both blocks are identical.

        Parameters
        ----------
        block : design.Block

        Returns
        -------
        identical: boolean

        See Also
        --------
        expyriment.design.Trial.compare

        """

        return (self.factor_dict == block.factor_dict)

    def get_random_trial(self):
        """Returns a randomly selected trial.

        Notes
        -----
        This function is useful for compiling training blocks.

        Returns
        -------
        rnd : design.Trial
            random Expyriment trial

        """

        rnd = rand_int(0, len(self._trials) - 1)
        return self._trials[rnd]

    def add_trial(self, trial, copies=1, random_position=False):
        """Adds one or multiple copies of the trial to the block.

        Parameters
        ----------
        trial : design.Trial
            trial to add
        copies : int, optional
            number of copies to add (default = 1)
        random_position : bool, optional
            True  = insert trials at random position;
            False = append trials at the end (default=False)

        """

        for _x in range(0, copies):
            if random_position:
                pos = rand_int(0, len(self._trials))
                self._trials.insert(pos, trial.copy())
            else:
                self._trials.append(trial.copy())
            self._trials[-1]._id = self._trial_id_counter
            self._trial_id_counter += 1

        log_txt = "Block,trial added,{0}, {1}".format(self.name,
                                                      self._trials[-1]._id)
        if random_position:
            log_txt = log_txt + ", random position"
        _internals.active_exp._event_file_log(log_txt, 2)

    def remove_trial(self, position):
        """Remove a trial.

        Parameters
        ----------
        position : int
            position of the trial

        """

        trial = self._trials.pop(position)

        _internals.active_exp._event_file_log(
            "Block,trial removed,{0},{1}".format(self.id, trial.id), 2)

    def clear_trials(self):
        """Clear all trials."""

        self._trials = []
        self._trial_id_counter = 0

        _internals.active_exp._event_file_log("Block,trials cleared", 2)

    @property
    def n_trials(self):
        """Getter for n_trials.

        Number of trials.

        """

        return len(self._trials)

    @property
    def trial_factor_names(self):
        """Getter for trial_factor_names.

        Get all factor names defined in trial list.

        """

        if len(self.trials) < 1:
            return []
        rtn = self.trials[0].factor_names
        for tr in self.trials:
            for new_fac in tr.factor_names:
                is_new = True
                for old_fac in rtn:
                    if old_fac == new_fac:
                        is_new = False
                        break
                if is_new:
                    rtn.append(new_fac)
        return rtn

    def get_trial_factor_values(self, name):
        """Return a list of the values of a certain factor for all trials.

        Parameters
        ----------
        name : str
            name of the factor

        """

        rtn = []
        for trial in self.trials:
            rtn.append(trial.get_factor(name, return_none_if_not_defined=True))
        return rtn

    @property
    def design_as_text(self):
        """Getter for design_as_text.

        List of trial factors as csv table.

        The list considers only the defined trial factors and not the
        added stimuli.

        """

        rtn = "{0},{1}".format(self._trial_cnt_variable_name,
                               self._trial_id_variable_name)
        factors = self.trial_factor_names
        for f in factors:
            rtn = rtn + ",{0}".format(f)
        for cnt, tr in enumerate(self.trials):
            rtn = rtn + "\n{0},{1}".format(cnt, tr.id)
            for f in factors:
                rtn = rtn + ",{0}".format(tr.get_factor(f, return_none_if_not_defined=True))
        return rtn

    def save_design(self, filename):
        """Save the list of trials to a csv file.

        The function considers only the defined trial factors and not the
        added stimuli.

        Parameters
        filename  -- name (fullpath) of the csv file

        """

        with open(filename, 'wb') as f:
            try:
                locale_enc = locale.getdefaultlocale()[1]
            except:
                locale_enc = "UTF-8"
            header = "# -*- coding: {0} -*-\n".format(locale_enc)
            f.write(unicode2byte(header + self.design_as_text))

    def read_design(self, filename):
        """Reads a list of trials from a csv file and clears the old block
        design.

        The function considers only the trial factors and not the added
        stimuli. All factors will be read in as text strings and not casted to
        numericals. Please do this manually if required.

        Parameters
        ----------
        filename : str
            name (fullpath) of the csv file (str)

        """

        tmp = Block()
        tmp.add_trials_from_csv_file(filename)
        fac_names = tmp.trial_factor_names

        self.clear_factors()
        self.clear_trials()

        for tr in tmp.trials:
            new = Trial()
            for fac in fac_names:
                if fac == self._trial_cnt_variable_name:
                    pass
                elif fac == self._trial_id_variable_name:
                    new._id = int(tr.get_factor(fac, return_none_if_not_defined=True))
                else:
                    new.set_factor(fac, tr.get_factor(fac, return_none_if_not_defined=True))
            self.add_trial(new)

    def add_trials_from_csv_file(self, filename, encoding=None):
        """Read a list of trials from csv-file and append the new trials to the
        block. Trials are defined as combinations of trial factors.

        **csv-file specifications**

            The first row of the csv-file specifies the factor names. Each
            following row describes one trial. Each row must have the same
            amount of columns.

        Notes
        ------
        All factors will be read in as text strings and not casted to
        numericals. Please do this manually if required.

        Parameters
        ----------
        filename : str
            name (fullpath) of the csv file (str)

        encoding : str, optional
            the encoding to be used when reading from the file

        """

        factor_names = []
        if encoding is None:
            with open(filename, 'r') as fl:
                first_line = fl.readline()
                encoding = re.findall("coding[:=]\s*([-\w.]+)", first_line)
                if encoding == []:
                    second_line = fl.readline()
                    encoding = re.findall("coding[:=]\s*([-\w.]+)",
                                          second_line)
                    if encoding == []:
                        if _internals.PYTHON3:
                            encoding = ['utf-8']
                        else:
                            encoding = [None]
        else:
            encoding = [encoding]

        with codecs.open(filename, "rb", encoding[0], errors='replace') as f:
            reader = csv.reader(f)
            for r_cnt, row in enumerate(reader):
                if r_cnt == 0:
                    factor_names = [byte2unicode(x) for x in row]
                else:
                    trial = Trial()
                    for c_cnt in range(0, len(row)):
                        trial.set_factor(byte2unicode(factor_names[c_cnt]),
                                         byte2unicode(row[c_cnt]))
                    self.add_trial(trial)


    def add_trials_full_factorial(self, design_dict, copies=1):
        """Add trials of all combinations of a full factorial design.

         Factors will be defined according the design_dict parameter.
         The design is specified by a dictionary. The keys of the dictionary
         indicate the factor names. Each value of the dictionary has to be
         a list of the factor levels.

         Parameters
         ----------
         design_dict : dictionary
             keys: factor names, values: lists of factor levels
         copies : int, optional
             number of copies to add (default = 1)

         Examples
         --------
         >>> bl = design.Block()
         >>> design = {
         >>>    "target": ["left", "center", "right"],
         >>>    "cue": [-300, 300],
         >>>    "letter": ["H", "F"]}
         >>> bl.add_trials_full_factorial(design, copies=10)

        """

        keys = design_dict.keys()
        levels = list(map(lambda x: len(design_dict[x]), keys))
        cnt = [0]*len(design_dict)

        while cnt is not None:
            tr = Trial()
            for i,k in enumerate(keys):
                tr.set_factor(k, design_dict[ k ][ cnt[i] ])
            self.add_trial(tr, copies=copies)
            cnt = _get_next_permutation(values=cnt, levels=levels)


    def order_trials(self, order):
        """Order the trials.

        Parameters
        ----------
        order : list
            list with the new order of positions

        """

        if not len(order) == len(self._trials):
            raise ValueError("Given order has wrong number of items!")
        trials_new = []
        for position in order:
            trials_new.append(self._trials[position])
        self._trials = trials_new

    def swap_trials(self, position1, position2):
        """Swap two trials.

        Parameters
        ----------
        position1 : int
            position of first trial
        position2 : int
            position of second trial

        """

        if position1 < len(self._trials) and position2 < len(self._trials):
            self._trials[position1], self._trials[position2] = \
                self._trials[position2], self._trials[position1]
            return True
        else:
            return False

    @property
    def max_trial_repetitions(self):
        """Getter for max_trial_repetitions.

        Returns the maximum number of immediate trial repetitions.

        """

        tmp = []
        for t in self._trials:
            tmp.append(t.factors_as_text)

        max_reps = 0
        cnt = 0
        for x in range(1, len(tmp) - 1):
            if tmp[x - 1] == tmp[x]:
                cnt += 1
                if cnt > max_reps:
                    max_reps = cnt
            else:
                cnt = 0
        return max_reps

    def shuffle_trials(self, method=0, max_repetitions=None,
                       n_segments=None):
        """Shuffle all trials.

        The function returns False if no randomization could be found that
        fulfills the max immediate trial repetition criterion. The different
        type of trials are only defined by the factors. Shuffle does not
        take into account the added stimuli.

        The following randomization methods are defined:

                0 = total randomization of trial order (default)

                1 = randomization within small miniblocks. Each miniblock
                contains one trial of each type (only defined by factors!);
                in other words, copies of one trial type are always in
                different miniblocks

        Parameters
        ----------
        method : int, optional
            method of trial randomization (default=0)
        max_repetitions : int, optional
            see documentation of `randomize.shuffle_list` (default = None)
        n_segments : int, optional
            this parameter will be only considered for randomization method 0;
            see documentation of `randomize.shuffle_list` (default = None)

        Returns
        -------
        succeeded : bool

        """

        if method == 1: # make segments
            tmp = self._trials
            self._trials = []
            types_occured = []
            cnt = 0
            n_segments = 1
            while len(tmp) > 0:
                is_new = True
                tr_type = tmp[cnt].factors_as_text
                for occ in types_occured:
                    if tr_type == occ:
                        is_new = False
                        break
                if is_new:
                    self._trials.append(tmp.pop(cnt))
                    types_occured.append(tr_type)
                    cnt = 0
                else:
                    cnt = cnt + 1
                    if cnt >= len(tmp):
                        types_occured = []
                        cnt = 0
                        if len(tmp)>0:
                            n_segments += 1

        rtn = shuffle_list(self._trials,
                                   max_repetitions=max_repetitions,
                                   n_segments=n_segments)
        if rtn == False:
            print("Warning: Could not find an appropriate trial " + \
                          "randomization!")
        return rtn

    def sort_trials(self):
        """Sort the trials according to their indices from low to high."""

        trials_new = []
        id_list = [x.id for x in self._trials]
        id_list.sort()
        for _id in id_list:
            position = [i for i, x in enumerate(self._trials)
                        if x.id == _id][0]
            trials_new.append(self._trials[position])
        self._trials = trials_new

    def find_trial(self, id):
        """Find the positions of a trial.

        Parameters
        -----------
        id : int
            trial id to look for

        Returns
        -------
        pos : list
            positions as a list or None if not in trial list.

        """

        positions = [i for i, x in enumerate(self._trials) if x.id == id]
        if positions:
            return positions

    def copy(self):
        """Return a copy of the block."""

        owntrials = []
        triallist = []
        for trial in self._trials:
            owntrials.append(trial)
            triallist.append(trial.copy())
        self._trials = None
        rtn = deepcopy(self)
        self._trials = owntrials
        rtn._trials = triallist
        return rtn


class Trial(object):
    """A class implementing an experimental trial."""

    def __init__(self):
        """Create a Trial."""

        self._stimuli = []
        self._factors = {}
        self._id = None

    @property
    def stimuli(self):
        """Getter for stimuli."""

        return self._stimuli

    @property
    def id(self):
        """Getter for id."""

        return self._id

    def __str__(self):
        return """
        Trial:   {0}

        Stimuli: {1}
        """.format(str(self.id),
                   [stimulus.id for stimulus in self.stimuli])

    def set_factor(self, name, value):
        """Set a factor for the trial.

        Parameters
        ----------
        name : str
            factor name
        value : str or numeric
            factor value

        """

        if isinstance(value, (bytes, str, int, float)):
            self._factors[name] = value
        else:
            message = "Factor values or factor conditions must to be a " + \
                "string or a numeric (i.e. float or integer).\n " + \
                "{0} is not allowed.".format(type(value))
            raise TypeError(message)

    def has_factor(self, name):
        """Checks if a factor is defined.

        Parameters
        ----------
        name : str
            factor name (str)

        Returns
        -------
        bool : boolean

        """

        return name in self.factor_names

    def get_factor(self, name, return_none_if_not_defined=False):
        """Get a factor of the trial.

        Parameters
        ----------
        name : str
            factor name
        return_none_if_not_defined : boolean (default: False)
            suppresses exception and returns None if factor is not defined.

        Returns
        -------
        factor_val : str or numeric

        """

        try:
            rtn = self._factors[name]
        except:
            if return_none_if_not_defined:
                return None
            else:
                raise RuntimeError(_FACTOR_NOT_EXIST.format(name))
        return rtn

    @property
    def factor_dict(self):
        """The dictionary with all factors of the trial."""

        return self._factors

    def clear_factors(self):
        """Clear all factors."""

        self._factors = {}

    @property
    def factor_names(self):
        """Getter for factors names."""

        return list(self._factors.keys())

    @property
    def factors_as_text(self):
        """Return all factor names and values as csv string line"""
        all_factors = ""
        for f in self.factor_names:
            all_factors = all_factors + u"{0}={1}, ".format(
                f, str(self.get_factor(f, return_none_if_not_defined=True)))
        return all_factors

    def compare(self, trial):
        """Compares this trial with another trail and returns `True` if all
        factors associated with both trials are identical. Added stimuli will
        be ignored for the comparison.

        Parameters
        ----------
        trail : design.Trial

        Returns
        -------
        identical: boolean

        """

        return (self.factor_dict == trial.factor_dict)

    def add_stimulus(self, stimulus):
        """Add a stimulus to the trial.

        Notes
        -----
        This will add references to stimuli, not copies!

        Parameters
        ----------
        stimulus : expyriment stimulus
            stimulus to add (expyriment.stimuli.* object)

        """

        self._stimuli.append(stimulus)

        _internals.active_exp._event_file_log(
            "Trial,stimulus added,{0},{1}".format(self.id, stimulus.id), 2)

    def remove_stimulus(self, position):
        """Remove stimulus from trial.

        Parameters
        ----------
        position : int
            position of the stimulus

        """

        stimulus = self._stimuli.pop(position)

        _internals.active_exp._event_file_log(
            "Trial,stimulus removed,{0},{1}".format(self.id, stimulus.id), 2)

    def order_stimuli(self, order):
        """Order the stimuli.

        Parameters
        ----------
        order : list
            list with the new order of positions

        """

        if not len(order) == len(self._stimuli):
            raise ValueError("Given order has wrong number of items!")
        stimuli_new = []
        for position in order:
            stimuli_new.append(self._stimuli[position])
        self._stimuli = stimuli_new

    def clear_stimuli(self):
        """Clear the stimuli."""

        self._stimuli = []
        _internals.active_exp._event_file_log("Trial,stimuli cleared", 2)

    def swap_stimuli(self, position1, position2):
        """Swap two stimuli.

        Parameters
        ----------
        position1 : int
            position of first stimulus
        position2 : int
            position of second stimulus

        """

        if position1 < len(self._stimuli) and position2 < len(self._stimuli):
            self._stimuli[position1], self._stimuli[position2] = \
                self._stimuli[position2], self._stimuli[position1]
            return True
        else:
            return False

    def shuffle_stimuli(self, max_repetitions=None, n_segments=None):
        """Shuffle all stimuli.

        Parameters
        ----------
        max_repetitions : int, optional
            see documentation of `randomize.shuffle_list`, default = None
        n_segments : int, optional
            see documentation of `randomize.shuffle_list`, default = None

        Returns
        -------
        success : bool
            randomization was successful and fulfilled the specified
            constrains (see max_repetitions)

        See Also
        ----------
        expyriment.design.randomize.shuffle_list

        """

        return shuffle_list(self.stimuli,
                               max_repetitions=max_repetitions,
                               n_segments=n_segments)

    def sort_stimuli(self):
        """Sort the stimuli according to their IDs from low to high."""

        stimuli_new = []
        id_list = [x.id for x in self._stimuli]
        id_list.sort()
        for _id in id_list:
            position = [i for i, x in enumerate(self._stimuli)
                        if x.id == _id][0]
            stimuli_new.append(self._stimuli[position])
        self._stimuli = stimuli_new

    def find_stimulus(self, id):
        """Find the positions of a stimulus.

        Parameters
        ----------
        id : int
            stimulus id to look for

        Returns
        -------
        pos : int
            positions as a list or None if not in stimuli list

        """

        positions = [i for i, x in enumerate(self._stimuli) if x.id == id]
        if positions:
            return positions

    def copy(self):
        """Return a copy of the trial."""

        stimlist = []
        for stim in self._stimuli:
            stimlist.append(stim)
        self._stimuli = None
        rtn = deepcopy(self)
        self._stimuli = rtn._stimuli = stimlist
        return rtn

    def preload_stimuli(self):
        """Preload all stimuli in trial.

        Returns
        -------
        time : int
            time it took to execute this method in ms

        """

        start = Clock.monotonic_time()
        for stim in self._stimuli:
            stim.preload()
        return int((Clock.monotonic_time() - start) * 1000)

    def unload_stimuli(self, keep_surface=False):
        """Unload all stimuli in trial.

        Parameters
        ----------
        keep_surface : bool, optional
            keep the surface after unloading (default = False)

        Returns
        -------
        time : int
            time it took to execute this method in ms

        """

        start = Clock.monotonic_time()
        for stim in self._stimuli:
            stim.unload(keep_surface=keep_surface)
        return int((Clock.monotonic_time() - start) * 1000)


def _get_next_permutation(values, levels):
    """helper function
    count the array of values up,
    levels define the max levels per unit.

    Returns None is end reached
    """

    if values is None:
        return [0]*len(levels)
    p = 0
    while True:
        values[p] += 1
        if values[p]>=levels[p]:
            values[p] = 0
            p  += 1
            if p>=len(values):
                return None
        else:
            return values





"""
Video playback.

This module contains a class implementing video playback.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os
import time
from types import FunctionType

import pygame
try:
    import android.mixer as mixer
except:
    import pygame.mixer as mixer

from . import defaults
from . import _visual
from ..misc import unicode2byte, Clock, has_internet_connection, which
from .._internals import CallbackQuitEvent
from .. import _internals


class Video(_visual.Stimulus):
    """A class implementing a general video stimulus.

    This class uses a background thread for playing the video!

    Notes
    -----
    When the backend is set to ``"pygame"``, only MPEG-1 videos with MP3 audio
    are supported. You can use ffmpeg (www.ffmpeg.org) to convert from other
    formats:

        ffmpeg -i <inputfile> -vcodec mpeg1video -acodec libmp3lame -intra -qscale 2  <outputfile.mpg>

    The -qscale option is the quality setting. It can take values from 1 to 31.
    1 is the best quality, but big file size. 31 is the worst quality, but
    small file size. Play around with this setting to get a good balance
    between quality and file size.

    When the audio from the video should be played as well, the audiosystem
    has to be stopped (by calling ``expyriment.control.stop_audiosystem()``)
    BEFORE the video stimulus is preloaded! After the stimulus has been played
    the audiosystem can be started again (by calling
    ``expyriment.control.start_audiosystem()``).

    When showing videos in large dimensions, and your computer is not fast
    enough, frames might be dropped! When using ``Video.wait_frame()`` or
    ``Video.wait_end()``, dropped video frames will be reported and logged.

    """

    @staticmethod
    def get_ffmpeg_binary():
        try:
            import imageio
            try:
                ffmpeg_binary = imageio.plugins.ffmpeg.get_exe()
                if ffmpeg_binary == "ffmpeg":
                    ffmpeg_binary = which(ffmpeg_binary)
                return ffmpeg_binary
            except imageio.core.NeedDownloadError:
                try:
                    assert has_internet_connection()
                    imageio.plugins.ffmpeg.download()
                    return imageio.plugins.ffmpeg.get_exe()
                except:
                    os.environ['IMAGEIO_NO_INTERNET'] = 'yes'
        except ImportError:
            pass

    def __init__(self, filename, backend=None, position=None):
        """Create a video stimulus.

        Parameters
        ----------
        filename : str
            filename (incl. path) of the video
        backend : str, optional
            'mediadecoder' or 'pygame'
        position : (int, int), optional
            position of the stimulus

        Notes
        -----
        When the backend is set to ``"pygame"``, only MPEG-1 videos with MP3
        audio are supported. You can use ffmpeg (www.ffmpeg.org) to convert from
        other formats:

            ffmpeg -i <inputfile> -vcodec mpeg1video -acodec libmp3lame -intra -qscale 2  <outputfile.mpg>

        The -qscale option is the quality setting. It can take values from 1 to 31.
        1 is the best quality, but big file size. 31 is the worst quality, but
        small file size. Play around with this setting to get a good balance
        between quality and file size.

        """

        from ..io import Keyboard

        _visual.Stimulus.__init__(self, filename)
        self.Keyboard = Keyboard()
        self._filename = filename
        self._is_preloaded = False
        self._is_paused = False
        self._frame = 0
        self._new_frame_available = False
        self._surface_locked = False
        self._audio_started = False
        if backend:
            self._backend = backend
        else:
            self._backend = defaults.video_backend
        if position:
            self._position = position
        else:
            self._position = defaults.video_position

        if not(os.path.isfile(self._filename)):
            raise IOError(u"The video file {0} does not exists".format(
                self._filename))

        if self._backend == "mediadecoder":
            Video.get_ffmpeg_binary()  # in case it still needs to be downloaded
            try:
                import mediadecoder as _mediadecoder
            except ImportError:
                print("Warning: Package 'mediadecoder' not installed!\n" +
                      "Video backend will be set to 'pygame'.")
                self._backend = "pygame"
            try:
                import sounddevice as _sounddevice
            except ImportError:
                print("Warning: Package 'sounddevice' not installed!\n" +
                      "Audio will be played back using Pygame audiosystem.")


    def __del__(self):
        """Destructor for the video stimulus."""

        self._surface = None
        self._file = None

    _getter_exception_message = "Cannot set {0} if preloaded!"

    @property
    def is_preloaded(self):
        """Getter for is_preloaded."""

        return self._is_preloaded

    @property
    def position(self):
        """Getter for position."""

        return self._position

    @property
    def filename(self):
        """Getter for filename."""

        return self._filename

    @filename.setter
    def filename(self, value):
        """Setter for filename."""

        if self._is_preloaded:
            raise AttributeError(Video._getter_exception_message.format(
                "filename"))
        else:
            self._filename = value
            if not(os.path.isfile(self._filename)):
                raise IOError(u"The video file {0} does not exists".format(
                    self._filename))

    @property
    def is_playing(self):
        """Property to check if video is playing."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.get_busy()
            elif self._backend == "mediadecoder":
                return self._file.status == 3

    @property
    def is_paused(self):
        """Property to check if video is paused."""

        if self._is_preloaded:
            return self._is_paused

    @property
    def time(self):
        """Property to get the current playback time."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.get_time()
            elif self._backend == "mediadecoder":
                return self._file.current_playtime

    @property
    def size(self):
        """Property to get the resolution of the video."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.get_size()
            elif self._backend == "mediadecoder":
                return self._file.clip.size

    @property
    def frame(self):
        """Property to get the current available video frame."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.get_frame()
            elif self._backend == "mediadecoder":
                return self._file.current_frame_no

    @property
    def new_frame_available(self):
        """Property to check if new video frame is available to render."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self.frame > self._frame
            elif self._backend == "mediadecoder":
                time.sleep(0.0001)  # Needed for performance for some reason
                return self._new_frame_available

    @property
    def length(self):
        """Property to get the length of the video."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.get_length()
            elif self._backend == "mediadecoder":
                return self._file.duration

    @property
    def has_video(self):
        """Property to check if video fifle has video information."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.has_video()
            elif self._backend == "mediadecoder":
                return self._file.clip.video is not None

    @property
    def has_audio(self):
        """Property to check if video file has audio information."""

        if self._is_preloaded:
            if self._backend == "pygame":
                return self._file.has_audio()
            elif self._backend == "mediadecoder":
                return self._file.clip.audio is not None

    def preload(self):
        """Preload stimulus to memory.

        Notes
        -----
        When the audio from the video should be played as well, the audiosystem
        has to be stopped (by calling ``expyriment.control.stop_audiosystem()``)
        BEFORE the video stimulus is preloaded! After the stimulus has been
        played the audiosystem can be started again (by calling
        ``expyriment.control.start_audiosystem()``).

        """

        if not self._is_preloaded:
            if not _internals.active_exp.is_initialized:
                message = "Can't preload video. Expyriment needs to be " + \
                          "initialized before preloading a video."
                raise RuntimeError(message)
            if self._backend == "pygame":
                self._file = pygame.movie.Movie(self._filename)
                size = self._file.get_size()
            elif self._backend == "mediadecoder":
                if self.get_ffmpeg_binary() is None:
                    raise RuntimeError("'ffmpeg' not found!")
                from mediadecoder.states import PLAYING
                from mediadecoder.decoder import Decoder
                self._file = Decoder(mediafile=self._filename,
                                     videorenderfunc=self._update_surface)
                if _internals.active_exp._screen.open_gl:
                    import moviepy.video.fx.all as vfx
                    self._file.clip = vfx.mirror_y(self._file.clip)
                if self._file.audioformat:
                    try:
                        from mediadecoder.soundrenderers.sounddevicerenderer import SoundrendererSounddevice
                        self._audio = SoundrendererSounddevice(
                            self._file.audioformat)
                        self._audio_renderer = "sounddevice"
                    except ImportError:
                        from mediadecoder.soundrenderers import SoundrendererPygame
                        self._audio = SoundrendererPygame(
                            self._file.audioformat)
                        self._audio_renderer = "pygame"
                    self._file.set_audiorenderer(self._audio)
                size = self._file.clip.size
            else:
                raise RuntimeError("Unknown backend '{0}'!".format(
                    self._backend))
            screen_size = _internals.active_exp.screen.surface.get_size()

            self._pos = [screen_size[0] // 2 - size[0] // 2 +
                         self._position[0],
                         screen_size[1] // 2 - size[1] // 2 -
                         self._position[1]]
            if self._backend== "pygame":
                self._surface = pygame.surface.Surface(size)
                self._file.set_display(self._surface)
            self._is_paused = False
            self._is_preloaded = True

    def unload(self, **kwargs):
        """Unload stimulus from memory.

        This removes the reference to the object in memory.
        It is up to the garbage collector to actually remove it from memory.

        """

        if self._is_preloaded:
            self._file = None
            self._surface = None
            self._is_preloaded = False

    def play(self, loop=False, log_event_tag=None, audio=True):
        """Play the video stimulus from the current position.

        Parameters
        ----------
        loop : bool, optional
            loop video playback (will be ignored when using play to unpause!)
        log_event_tag : numeral or string, optional
            if log_event_tag is defined and if logging is switched on for this
            stimulus (default), a summary of the inter-event-intervalls are
            appended at the end of the event file
        audio : bool, optional
            whether audio of video (if present) should be played (default=True)

        Notes
        -----
        When the audio from the video should be played as well, the audiosystem
        has to be stopped (by calling expyriment.control.stop_audiosystem() )
        BEFORE the video stimulus is preloaded! After the stimulus has been
        played the audiosystem can be started again (by calling
        expyriment.control.start_audiosystem() ).

        When showing videos in large dimensions, and your computer is not fast
        enough, frames might be dropped! When using Video.wait_frame() or
        Video.wait_end(), dropped video frames will be reported and logged.

        """

        if self._is_paused:
            if audio and not self._audio_started:
                self._audio.stream.start()
                self._audio_started = True
            elif not audio and self._audio_started:
                self._audio.stream.stop()
                self._audio_started = False
            self.pause()
        else:
            if self._backend == "pygame" and mixer.get_init() is not None:
                message = "Mixer is still initialized, cannot play audio! Call \
    expyriment.control.stop_audiosystem() before preloading the video."
                print("Warning: ", message)
                if self._logging:
                    _internals.active_exp._event_file_log(
                        "Video,warning," + message)

            if not self._is_preloaded:
                self.preload()
            if self._logging:
                _internals.active_exp._event_file_log(
                    "Video,playing,{0}".format(self._filename),
                    log_level=1, log_event_tag=log_event_tag)
            if self._backend == "mediadecoder" and self._file.audioformat and audio:
                self._audio.start()
                self._audio_started = True
            self._file.loop = loop
            self._file.play()

    def stop(self):
        """Stop the video stimulus."""

        if self._is_preloaded:
            self._file.stop()
            self.rewind()
            if self._backend == "mediadecoder" and self._file.audioformat:
                self._audio.close_stream()

    def pause(self):
        """Pause the video stimulus."""

        if self._is_preloaded:
            self._file.pause()
            if self._is_paused:
                self._is_paused = False
            else:
                self._is_paused = True

    def forward(self, seconds):
        """Advance playback position.

        Parameters
        ----------
        seconds : int
            amount to advance (in seconds)

        Notes
        -----
        When the backend is set to ``"pygame"``, this will not forward
        immediately, but play a short period of the beginning of the file! This
        is a Pygame issue which we cannot fix right now.

        """

        if self._is_preloaded:
            if self._backend == "pygame":
                self._file.skip(float(seconds))
                new_frame = self._file.get_frame()
            else:
                pos = self._file.current_playtime + seconds
                self._file.seek(pos)
                new_frame = int(pos * self._file.fps)
            self._frame = new_frame

    def rewind(self):
        """Rewind to start of video stimulus."""

        if self._is_preloaded:
            self._file.rewind()
            self._frame = 0

    def _update_surface(self, frame):
        """Update surface with newly available video frame."""

        if self._backend == "mediadecoder" and not self._surface_locked:
            self._surface = frame
            self._new_frame_available = True

    def present(self):
        """Present current frame.

        This method waits for the next frame and presents it on the screen.
        When using OpenGL, the method blocks until this frame is actually being
        written to the screen.

        Notes
        -----
        When the audio from the video should be played as well, the audiosystem
        has to be stopped (by calling ``expyriment.control.stop_audiosystem()``)
        BEFORE the video stimulus is preloaded! After the stimulus has been
        played the audiosystem can be started again (by calling
        ``expyriment.control.start_audiosystem()``).

        When showing videos in large dimensions, and your computer is not fast
        enough, frames might be dropped! When using ``Video.wait_frame()`` or
        ``Video.wait_end()``, dropped video frames will be reported and logged.

        """

        start = Clock.monotonic_time()
        while not self.new_frame_available:
            if not self.is_playing:
                return
        diff = self.frame - self._frame
        if diff > 1:
            warn_message = repr(diff - 1) + " video frame(s) dropped!"
            print(warn_message)
            _internals.active_exp._event_file_warn(
                "Video,warning," + warn_message)
        self._frame = self.frame
        self.update()
        return (Clock.monotonic_time() - start) * 1000

    def update(self):
        """Update the screen."""

        if not self.is_playing:
            return
        start = Clock.monotonic_time()
        self._surface_locked = True
        if not _internals.active_exp._screen.open_gl:
            _internals.active_exp._screen.surface.blit(
                pygame.surfarray.make_surface(self._surface.swapaxes(0,1)),
                self._pos)
            self._surface_locked = False
            self._new_frame_available = False
        else:
            ogl_screen = _visual._LaminaPanelSurface(
                self._surface, quadDims=(1,1,1,1), position=self._position)
            self._surface_locked = False
            self._new_frame_available = False
            ogl_screen.display()
        _internals.active_exp._screen.update()

    def _wait(self, frame=None, callback_function=None,
              process_control_events=True):
        """Wait until frame was shown or end of video and update screen.

        Parameters
        ----------
        frame : int, optional
            number of the frame to stop after
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Notes
        ------
        This will also by default process control events (quit and pause).
        Thus, keyboard events will be cleared from the cue and cannot be
        received by a Keyboard().check() anymore!

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        while self.is_playing:

            if _internals.skip_wait_methods:
                return

            self.present()
            if frame is not None and self.frame >= frame:
                break

            if isinstance(callback_function, FunctionType):
                rtn_callback = callback_function()
                if isinstance(rtn_callback, CallbackQuitEvent):
                    return rtn_callback
            if _internals.active_exp.is_initialized:
                rtn_callback = _internals.active_exp._execute_wait_callback()
                if isinstance(rtn_callback, CallbackQuitEvent):
                    return rtn_callback
                if process_control_events:
                    if _internals.active_exp.mouse.process_quit_event():
                        break

            for event in pygame.event.get(pygame.KEYDOWN):
                if _internals.active_exp.is_initialized and \
                   process_control_events and \
                   event.type == pygame.KEYDOWN and (
                       event.key == self.Keyboard.get_quit_key() or
                       event.key == self.Keyboard.get_pause_key()):
                    self.pause()
                    self.Keyboard.process_control_keys(event, self.stop)
                    self.play()

    def wait_frame(self, frame, callback_function=None,
                   process_control_events=True):
        """Wait until certain frame was shown and constantly update screen.

        Parameters
        ----------
        frame : int
            number of the frame to stop after
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Notes
        ------
        This will also by default process control events (quit and pause).
        Thus, keyboard events will be cleared from the cue and cannot be
        received by a Keyboard().check() anymore!
        If keyboard events should not be cleared, a loop has to be created
        manually like::

            video.present()
            while video.is_playing and video.frame < frame:
                while not video.new_frame_available:
                    key = exp.keyboard.check()
                    if key == ...
                video.update()
            video.stop()

        """

        if self.is_playing:
            self._wait(frame, callback_function, process_control_events)

    def wait_end(self, callback_function=None, process_control_events=True):
        """Wait until video has ended and constantly update screen.

        Parameters
        ----------
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Notes
        ------
        This will also by default process control events (quit and pause).
        Thus, keyboard events will be cleared from the cue and cannot be
        received by a Keyboard().check() anymore!
        If keyboard events should not be cleared, a loop has to be created
        manually like::

            video.present()
            while video.is_playing and video.frame < frame:
                while not video.new_frame_available:
                    key = exp.keyboard.check()
                    if key == ...
                video.update()
            video.stop()

        """

        self._wait(callback_function, process_control_events)



"""
The miscellaneous module.

This module contains miscellaneous functions for expyriment.

All classes in this module should be called directly via expyriment.misc.*:

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

import os
import sys
import glob
import random
import colorsys

import pygame

from .._internals import PYTHON3, android, get_settings_folder, get_version

try:
    from locale import getdefaultlocale
    LOCALE_ENC = getdefaultlocale()[1]
except ImportError:
    LOCALE_ENC = None  # Not available on Android

if LOCALE_ENC is None:
    LOCALE_ENC = 'utf-8'

FS_ENC = sys.getfilesystemencoding()
if FS_ENC is None:
    FS_ENC = 'utf-8'


_round = round

def round(number, ndigits=0):
    """Round half away from zero.

    This method implements the Python 2 way of rounding.
    For "bankers rounding" (round half to even), plese use the builtin `round`
    function in Python 3 or Numpy's `around` function in Python 2.

    Parameters
    ----------
    number : int or float
        the number to be rounded
    ndigits : int
        the number of digits to round to (default = 0)

    Returns
    -------
    rounded_number : float
        the rounded number

    """

    import math

    if not PYTHON3:
        return _round(number, ndigits)
    else:
        p = 10 ** ndigits
        if number > 0:
            return float(math.floor((number * p) + 0.5))/p
        else:
            return float(math.ceil((number * p) - 0.5))/p


def compare_codes(input_code, standard_codes, bitwise_comparison=True):
    """Helper function to compare input_code with a standard codes.

    Returns a boolean and operates by default bitwise.

    Parameters
    ----------
    input_code : int
        code or bitpattern
    standard_codes : int or list
        code/bitpattern or list of codes/bitpattern
    bitwise_comparison : bool, optional
        (default = True)

    """

    if isinstance(standard_codes, (list, tuple)):
        for code in standard_codes:
            if compare_codes(input_code, code, bitwise_comparison):
                return True
        return False
    else:
        if input_code == standard_codes:  # accounts also for (bitwise) 0==0 &
                                          # None==None
            return True
        elif bitwise_comparison:
            return (input_code & standard_codes)
        else:
            return False


def byte2unicode(s, fse=False):
    if (PYTHON3 and isinstance(s, str)) or\
       (not PYTHON3 and isinstance(s, unicode)):
        return s

    if fse:
        try:
            u = s.decode(FS_ENC)
        except UnicodeDecodeError:
            u = s.decode('utf-8', 'replace')
    else:
        try:
            u = s.decode(LOCALE_ENC)
        except UnicodeDecodeError:
            u = s.decode('utf-8', 'replace')
    return u


def unicode2byte(u, fse=False):
    if isinstance(u, bytes):
        return u

    if fse:
        try:
            s = u.encode(FS_ENC)
        except UnicodeEncodeError:
            s = u.encode('utf-8', 'replace')
    else:
        try:
            s = u.encode(LOCALE_ENC)
        except UnicodeEncodeError:
            s = u.encode('utf-8', 'replace')
    return s


def str2unicode(s, fse=False):
    """Convert str to unicode.

    Converts an input str or unicode object to a unicode object without
    throwing an exception. If fse is False, the first encoding that is tried is
    the encoding according to the locale settings, falling back to utf-8
    encoding if this throws an error. If fse is True, the filesystem encoding
    is tried, falling back to utf-8. Unicode input objects are returned
    unmodified.

    Parameters
    ----------
    s : str or unicode
        input text
    fse : bool
        indicates whether the filesystem encoding should be tried first.
        (default = False)

    Returns
    -------
    A unicode-type string.
    """
    return byte2unicode(s, fse)

def unicode2str(u, fse=False):
    """Convert unicode to str.

    Converts an input str or unicode object to a str object without throwing
    an exception. If fse is False, the str is encoded according to the locale
    (with utf-8 as a fallback), otherwise it is encoded with the
    filesystemencoding. Str input objects are returned unmodified.

    Parameters
    ----------
    u : str or unicode
    input text
    fse : bool
    indicates whether the filesystem encoding should used.
    (default = False)

    Returns
    -------
    A str-type string.
    """

    return unicode2byte(u, fse)

def numpad_digit_code2ascii(keycode):
    """Convert numpad keycode to the ascii code of that particular number

    If it is not a keypad digit code, no conversion takes place and the
    same code will be returned.

    Returns
    -------
    ascii_code : int

    """

    from ..misc import constants
    if keycode in constants.K_ALL_KEYPAD_DIGITS:
        return keycode - (constants.K_KP1 - constants.K_1)
    else:
        return keycode

def add_fonts(folder):
    """Add fonts to Expyriment.

    All truetype fonts found in the given folder will be added to
    Expyriment, such that they are found when only giving their name
    (i.e. without the full path).

    Parameters
    ----------
    folder : str or unicode
        the full path to the folder to search for

    """

    # If font cache has to be (re-)created, initializing system fonts can take
    # a while. By having a watchdog thread, we can check if this is the case
    # and notify the user accordingly.

    import time
    import threading

    def watchdog_timer(state):
        time.sleep(1)
        if not state['completed']:
            m = "Initializing system fonts. This might take a couple of minutes..."
            sys.stdout.write(m + '\n')

    state = {'completed': False}
    watchdog = threading.Thread(target=watchdog_timer, args=(state,))
    watchdog.daemon = True
    watchdog.start()
    pygame.font.init()
    pygame.sysfont.initsysfonts()
    state['completed'] = True

    for font in glob.glob(os.path.join(folder, "*")):
        if font[-4:].lower() in ['.ttf', '.ttc']:
            font = byte2unicode(font, fse=True)
            name = os.path.split(font)[1]
            bold = name.find('Bold') >= 0
            italic = name.find('Italic') >= 0
            oblique = name.find('Oblique') >= 0
            name = name.replace(".ttf", "")
            if name.endswith("Regular"):
                name = name.replace("Regular", "")
            if name.endswith("Bold"):
                name = name.replace("Bold", "")
            if name.endswith("Italic"):
                name = name.replace("Italic", "")
            if name.endswith("Oblique"):
                name = name.replace("Oblique", "")
            name = ''.join([c.lower() for c in name if c.isalnum()])
            pygame.sysfont._addfont(name, bold, italic or oblique, font,
                                    pygame.sysfont.Sysfonts)


def list_fonts():
    """List all fonts installed on the system.

    Returns a dictionary where the key is the font name and the value is the
    absolute path to the font file.

    """

    pygame.font.init()

    d = {}
    fonts = pygame.font.get_fonts()
    for font in fonts:
        d[font] = pygame.font.match_font(font)
    return d


def find_font(font):
    """Find an installed font given a font name.

    This will try to match a font installed on the system that is similar to
    the given font name.

    Parameters
    ----------
    font : str
        name of the font

    Returns
    -------
    font : str
        the font that is most similar
        If no font is found, an empty string will be returned.

    """

    pygame.font.init()

    if os.path.isfile(font):
        return font
    font_file = pygame.font.match_font(font)
    if font_file is not None:
        return font_file
    else:
        warn_message = "Failed to find font {0}!".format(font)
        print("Warning: " + warn_message)
        return ""


def get_monitor_resolution():
    """Returns the monitor resolution

    Returns
    -------
    resolution: (int, int)
        monitor resolution, screen resolution

    """

    from .. import _internals
    if _internals.active_exp.is_initialized:
        return _internals.active_exp.screen.monitor_resolution
    else:
        pygame.display.init()
        return (pygame.display.Info().current_w,
                pygame.display.Info().current_h)


def is_ipython_running():
    """Return True if IPython is running."""

    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def is_idle_running():
    """Return True if IDLE is running."""

    return "idlelib.run" in sys.modules


def is_interactive_mode():
    """Returns if Python is running in interactive mode (such as IDLE or
    IPthon)

    Returns
    -------
    interactive_mode : boolean

    """

    # ps2 is only defined in interactive mode
    return hasattr(sys, "ps2") or is_idle_running() or is_ipython_running()


def is_android_running():
    """Return True if Exypriment runs on Android."""

    return android is not None


def has_internet_connection():
    """Return True if computer is connected to internet."""
    try:
        import socket
        host = socket.gethostbyname("google.com")
        socket.create_connection((host, 80), 2)
        return True
    except:
        return False


def create_colours(amount):
    """Create different and equally spaced RGB colours.

    Parameters
    ----------
    amount : int
        the number of colours to create

    Returns
    -------
    colours : list
        a list of colours, each in the form [r, g, b]

    """

    colours = []
    for i in range(0, 360, 360//amount):
        h = i / 360.0
        l = (50 + random.random() * 10) / 100
        s = (90 + random.random() * 10) / 100
        colours.append([int(x * 255) for x in colorsys.hls_to_rgb(h, l, s)])
    return colours


def which(programme):
    """Locate a programme file in the user's path.

    This mimics behaviour of UNIX's 'which' command.

    Parameters
    ----------
    programme : str
        the programme to file to locate

    Returns
    -------
    path : str or None
        the full path to the programme file or None if not found

    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(programme)
    if fpath:
        if is_exe(programme):
            return programme
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, programme)
            if is_exe(exe_file):
                return exe_file

    return None


def download_from_stash(content="all", branch=None):
    """Download content from the Expyriment stash.

    Content will be stored in an Expyriment settings diretory (`.expyriment`
    or `~expyriment`, located in the current user's home directory).

    Parameters
    ----------
    content : str, optional
        the content to install ("examples", "extras", "tools", or "all")
        (default="all")
    branch : str, optional
        a specific branch to get the content from (if not set, the branch
        corresponding to the current Expyriment release is used)
        (default=None)

    """

    def show_progress(count, total, status=''):
        bar_len = 40
        filled_len = int(round(bar_len * count / float(total)))
        percents = round(100.0 * count / float(total), 1)
        bar = '=' * filled_len + ' ' * (bar_len - filled_len)
        sys.stdout.write('{:5.1f}% [{}] {}\r'.format(percents, bar, status))
        sys.stdout.flush()

    if PYTHON3:
        from urllib.request import urlopen, Request
    else:
        from urllib2 import urlopen

    from tempfile import TemporaryFile
    from zipfile import ZipFile
    from shutil import copyfileobj

    if branch is None:
        branch = get_version().split(" ")[0]
    url = "https://github.com/expyriment/expyriment-stash/archive/{0}.zip"
    url = url.format(branch)
    try:
        r = Request(url, headers={"Accept-Encoding": "gzip; deflate"})
        u = urlopen(r)
    except:
        raise RuntimeError("Download of {0} failed!".format(url))

    with TemporaryFile() as f:
        meta = u.info()
        try:
            if PYTHON3:
                file_size = int(u.getheader('Content-Length'))
            else:
                file_size = int(u.info().getheaders("Content-Length")[0])
            file_size_dl = 0
            block_sz = 8192
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break
                file_size_dl += len(buffer)
                f.write(buffer)
                show_progress(file_size_dl, file_size,
                              "downloading stash ({0})".format(branch))
            sys.stdout.write("\n")
        except:
            show_progress(0, 100,
                          "downloading stash ({0})".format(branch))
            chunk = u.read()
            while chunk:
                f.write(chunk)
                chunk = u.read()
            show_progress(100, 100,
                          "downloading stash ({0})".format(branch))

        if get_settings_folder() is None:
            os.makedirs(os.path.join(os.path.expanduser("~"), ".expyriment"))
        path = get_settings_folder()

        f_zip = ZipFile(f)
        check = f_zip.testzip()
        if check is not None:
            raise RuntimeError("Download was corrupted!")
        if content == "all":
            content_ = ("examples", "extras", "tools")
        else:
            content_ = content
        if not type(content_) in (list, tuple):
            content_ = (content_,)
        files = []
        root = f_zip.namelist()[0]
        for c in content_:
            mask = os.path.join(root, c)
            files.extend([x for x in f_zip.namelist() if x.startswith(mask)])
        files_installed = 0
        for member in files:
            filename = os.path.basename(member)
            if not filename:
                try:
                    os.mkdir(os.path.join(path, os.path.relpath(member, root)))
                except:
                    pass
                files_installed += 1
                continue
            source = f_zip.open(member)
            target = open(os.path.join(path, os.path.relpath(member, root)),
                          'wb')
            with source, target:
                copyfileobj(source, target)
            files_installed += 1
            show_progress(files_installed, len(files),
                          "installing content ({0})".format(content))
    print("")


def py2py3_sort_array(array):
    """Sorts an array with different types using the string representation
    under Python2 and Python3. Sorts in place!

    Returns
    -------
    array: the sorted array

    """

    array.sort(key=_sorter_fnc)
    return array

def _sorter_fnc(x):
    """sorter function for py_sort"""
    if x is None:
        return str("")
    else:
        return str(x)



"""
A TextMenu.

This module contains a class implementing a TextMenu.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

from . import defaults
from ._keyboard import Keyboard
from ._mouse import Mouse
from ._input_output import Input
from .. import stimuli, misc
from .. import _internals


class TextMenu(Input):
    """A class implementing a text menu."""

    def __init__(self, heading, menu_items, width, position=None,
                 text_size=None, gap=None, heading_font=None,
                 text_font=None, background_colour=None,
                 text_colour=None, heading_text_colour=None,
                 select_background_colour=None, select_text_colour=None,
                 select_frame_colour=None, select_frame_line_width=None,
                 justification=None, scroll_menu=None,
                 background_stimulus=None, mouse=None):
        """Create a text menu.

        This creates a menu with items to be selected.

        Parameters
        ----------
        heading : str
            menu heading
        menu_items : str or list
            list with menu items
        width : int
            width of the menu
        position : (int, int), optional
        text_size : int, optional
        background_colour : (int, int, int), optional
            background colour of the menu
        gap : int, optional
            size of gap (pixel) between heading and item list
        heading_font : str, optional
            font to be used for the heading
        text_font : str, optional
            font to be used for the text
        text_colour : (int, int, int), optional
            text colour of the items
        heading_text_colour : (int, int, int), optional
            text colour of the heading
        select_background_colour : (int, int, int), optional
            background colour of the currently selected item
        select_text_colour : (int, int, int), optional
            text colour of the currently selected item
        select_frame_colour : (int, int, int), optional
            colour of the frame around the selected item
        select_frame_line_width : int, optional
            line width of the frame around the selected item
        justification : int, optional
            text justification: 0 (left), 1 (center), 2 (right)
        scroll_menu : int, optional
            maximum length of a item list before a scroll menu will
            be display. If the parameter is 0 of False scroll menu
            will not be displayed
        background_stimulus : visual expyriment stimulus, optional
            The background stimulus is a second stimulus that will be presented
            together with the TextMenu. For both stimuli overlap TextMenu
            will appear on top of the background_stimulus
        mouse : expyriment.io.Mouse object or True, optional
            If True or a mouse object is given, the menu will be controlled
            by the mouse

        """

        if position is None:
            position = defaults.textmenu_position
        if gap is None:
            gap = defaults.textmenu_gap
        if heading_font is None:
            heading_font = defaults.textmenu_text_font
        if text_font is None:
            text_font = defaults.textmenu_text_font
        if text_size is None:
            text_size = defaults.textmenu_text_size
        if background_colour is None:
            background_colour = defaults.textmenu_background_colour
        if text_colour is None:
            text_colour = defaults.textmenu_text_colour
        if heading_text_colour is None:
            heading_text_colour = defaults.textmenu_heading_text_colour
        if select_background_colour is None:
            select_background_colour = \
                defaults.textmenu_select_background_colour
        if select_text_colour is None:
            select_text_colour = defaults.textmenu_select_text_colour
        if select_frame_colour is None:
            select_frame_colour = defaults.textmenu_select_frame_colour
        if select_frame_line_width is None:
            select_frame_line_width = defaults.textmenu_select_frame_line_width
        if justification is None:
            justification = defaults.textmenu_justification
        if scroll_menu is None:
            scroll_menu = defaults.textmenu_scroll_menu

        self._scroll_menu = abs(int(scroll_menu))
        if self._scroll_menu > 0 and self._scroll_menu < 5:
            self._scroll_menu = 5
        self._gap = gap
        self._position = position
        self._bkg_colours = [background_colour, select_background_colour]
        self._text_colours = [text_colour, select_text_colour]
        self._line_size = (width, stimuli.TextLine(
            menu_items[0], text_size=text_size).surface_size[1] + 2)
        stimuli._stimulus.Stimulus._id_counter -= 1
        self._frame = stimuli.Rectangle(
            line_width=select_frame_line_width,
            size=(self._line_size[0] + 2 * select_frame_line_width,
                  self._line_size[1] + 2 * select_frame_line_width),
            colour=select_frame_colour)
        stimuli._stimulus.Stimulus._id_counter -= 1
        if background_stimulus is not None:
            if background_stimulus.__class__.__base__ in \
                    [stimuli._visual.Visual, stimuli.Shape]:
                self._background_stimulus = background_stimulus
            else:
                raise TypeError("{0} ".format(type(background_stimulus)) +
                                "is not a valid background stimulus. " +
                                "Use an expyriment visual stimulus.")
        else:
            self._background_stimulus = None

        if mouse is True:
            self._mouse = _internals.active_exp.mouse
        elif isinstance(mouse, Mouse):
            self._mouse = mouse
        else:
            self._mouse = None

        self._canvas = stimuli.BlankScreen()
        stimuli._stimulus.Stimulus._id_counter -= 1
        self._original_items = menu_items
        self._menu_items = []
        for item in menu_items:
            self._menu_items.append(stimuli.TextBox(item,
                text_size=text_size, text_font=text_font,
                text_justification=justification,
                size=self._line_size))
            stimuli._stimulus.Stimulus._id_counter -= 1
        self._heading = stimuli.TextBox(
            heading,
            text_size=text_size,
            text_justification=justification,
            text_font=heading_font,
            text_colour=heading_text_colour,
            text_bold=True,
            background_colour=self._bkg_colours[0],
            size=self._line_size)
        stimuli._stimulus.Stimulus._id_counter -= 1

    @property
    def heading(self):
        """Getter for heading"""
        return self._heading

    @property
    def menu_items(self):
        """Getter for menu_items"""
        return self._menu_items

    @property
    def position(self):
        """Getter for position"""
        return self._position

    @property
    def text_size(self):
        """Getter for text_size"""
        return self._heading.text_size

    @property
    def background_colour(self):
        """Getter for background_colour"""
        return self._bkg_colours[0]

    @property
    def select_background_colour(self):
        """Getter for select_background_colour"""
        return self._bkg_colours[1]

    @property
    def gap(self):
        """Getter for gap"""
        return self._gap

    @property
    def text_colour(self):
        """Getter for text_colour"""
        return self._text_colours[0]

    @property
    def select_text_colour(self):
        """Getter for select_text_colour"""
        return self._text_colours[1]

    @property
    def heading_text_colour(self):
        """Getter for heading_text_colour"""
        return self._heading.text_colour

    @property
    def select_frame_colour(self):
        """Getter for select_frame_colour"""
        return self._frame.colour

    @property
    def select_frame_line_width(self):
        """Getter for select_frame_line_width"""
        return self._frame.line_width

    @property
    def justification(self):
        """Getter for justification"""
        return self._heading.text_justification

    @property
    def scroll_menu(self):
        """Getter for scroll_menu"""
        return self._scroll_menu

    @property
    def background_stimulus(self):
        """Getter for background_stimulus"""
        return self._background_stimulus

    def _append_item(self, item, is_selected, y_position):
        """helper function"""
        item.clear_surface()
        item.position = (self._position[0], y_position + self._position[1])
        if is_selected:
            item.background_colour = self._bkg_colours[1]
            item.text_colour = self._text_colours[1]
        else:
            item.background_colour = self._bkg_colours[0]
            item.text_colour = self._text_colours[0]
        item.plot(self._canvas)

    def _redraw(self, selected_item):
        """helper function"""
        if self._scroll_menu > 0:
            n = self._scroll_menu
        else:
            n = len(self._menu_items)
        self._canvas.clear_surface()
        if self._background_stimulus is not None:
            self._background_stimulus.plot(self._canvas)
        y_pos = int(((1.5 + n) * self._line_size[1]) + (n * self._gap)) // 2
        self._heading.position = (self._position[0], y_pos + self._position[1])
        self._heading.plot(self._canvas)
        y_pos = y_pos - int(0.5 * self._line_size[1])

        if self._scroll_menu == 0:
            for cnt, item in enumerate(self._menu_items):
                y_pos -= (self._line_size[1] + self._gap)
                self._append_item(item, cnt == selected_item, y_pos)
                if cnt == selected_item:
                    self._frame.position = (0, y_pos)
        else:  # scroll menu
            for cnt in range(selected_item - self._scroll_menu // 2,
                             selected_item + 1 + self._scroll_menu // 2):
                y_pos -= (self._line_size[1] + self._gap)
                if cnt >= 0 and cnt < len(self._menu_items):
                    self._append_item(self._menu_items[cnt],
                                      cnt == selected_item,
                                      y_pos)
                    if cnt == selected_item:
                        self._frame.position = (0, y_pos)

        if self._frame.line_width > 0:
            self._frame.plot(self._canvas)
        self._canvas.present()

    def get(self, preselected_item=0):
        """Present the menu and return the selected item.

        Parameters
        ----------
        preselected_item : int, optional
            item that is preselected when showing menu

        Returns
        -------
        selected : int
            integer representing the selected item in the list

        """

        selected = preselected_item
        # Keyboard
        if self._mouse is None:
            while True:
                self._redraw(selected)
                key = Keyboard().wait()[0]
                if key == misc.constants.K_UP:
                    selected -= 1
                elif key == misc.constants.K_DOWN:
                    selected += 1
                elif key in misc.constants.K_ALL_DIGITS and\
                        key > misc.constants.K_0:
                    selected = key - misc.constants.K_1
                elif key == misc.constants.K_RETURN:
                    break
                if selected < 0:
                    selected = 0
                elif selected >= len(self._menu_items):
                    selected = len(self._menu_items) - 1
            return selected
        # Mouse
        else:
            while True:
                pressed = None
                self._redraw(selected)
                event, pos, rt = self._mouse.wait_press()
                if self._scroll_menu == 0:
                    for cnt in range(len(self._menu_items)):
                        if 0 <= cnt < len(self._menu_items):
                            if self._menu_items[cnt].overlapping_with_position(pos):
                                pressed = cnt
                else:
                    for cnt in range(selected-self._scroll_menu // 2,
                                     selected+1+self._scroll_menu // 2):
                        if 0 <= cnt < len(self._menu_items):
                            if self._menu_items[cnt].overlapping_with_position(pos):
                                pressed = cnt
                if pressed is not None:
                    if pressed == selected and event <=2:
                        # selected and not mouse wheel
                        break
                    else:
                        selected = pressed
            return self._original_items[pressed]


    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging = 0
            exp_ = control.initialize()
        menu = TextMenu(heading="Expyriment TextMenu",
                        menu_items=["Items 1", "Items 1", "Items 3",
                                    "Items 4", "Items 5"],
                        width=250, select_frame_line_width = 5)
        print(menu.get())
